import textwrap

from bloqade import qasm2
from kirin.dialects import ilist
from bloqade.qasm2.dialects import glob, noise, inline, parallel


def test_inline():

    lines = textwrap.dedent(
        """
    OPENQASM 2.0;

    qreg q[2];
    creg c[2];

    h q[0];
    CX q[0], q[1];
    barrier q[0], q[1];
    CX q[0], q[1];
    rx(pi/2) q[0];
    """
    )

    @qasm2.main.add(inline)
    def qasm2_inline_code():
        qasm2.inline(lines)

    qasm2_inline_code.print()


def test_inline_ext():

    lines = textwrap.dedent(
        r"""
    KIRIN {qasm2.uop, qasm2.expr, qasm2.parallel, qasm2.glob, qasm2.noise};

    qreg q[3];
    creg c[3];

    h q[0];
    CX q[0], q[1];
    barrier q[0], q[1];
    CX q[0], q[1];
    rx(pi/2) q[0];
    glob.U(1.0, 2.0, 3.0) {q}
    noise.PAULI1(1.0, 2.0, 3.0) q[0];
    parallel.U(1.0, 2.0, 3.0) {q[0]; q[1]; q[2];}
    parallel.CZ {
    q[0], q[1];
    q[2], q[3];
    }
    parallel.RZ(2.1) {
        q[0];
        q[2];
    }
    """
    )
    print(lines)

    @qasm2.main.add(inline).add(parallel).add(glob).add(noise).add(ilist)
    def qasm2_inline_code():
        qasm2.inline(lines)

    qasm2_inline_code.print()


if __name__ == "__main__":
    # test_inline()

    lines = textwrap.dedent(
        """
    KIRIN {qasm2.glob, qasm2.uop};
    include "qelib1.inc";

    qreg q1[2];
    qreg q2[3];

    glob.U(1.0, 2.0, 3.0) {q1, q2}
    """
    )

    print(lines)

    @qasm2.extended.add(inline)
    def qasm2_inline_code():
        qasm2.inline(lines)

    qasm2_inline_code.print()



from typing import TypeVar

from kirin import ir, interp
from bloqade.types import QubitType
from kirin.analysis import Forward, const
from kirin.analysis.forward import ForwardFrame

from .lattice import Address


class AddressAnalysis(Forward[Address]):
    keys = ["qubit.address"]
    lattice = Address

    def initialize(self):
        super().initialize()
        self.next_address: int = 0
        return self

    @property
    def qubit_count(self) -> int:
        return self.next_address

    T = TypeVar("T")

    def get_const_value(self, typ: type[T], value: ir.SSAValue) -> T:
        if isinstance(hint := value.hints.get("const"), const.Value):
            data = hint.data
            if isinstance(data, typ):
                return hint.data
            raise interp.InterpreterError(
                f"Expected constant value <type = {typ}>, got {data}"
            )
        raise interp.InterpreterError(
            f"Expected constant value <type = {typ}>, got {value}"
        )

    def eval_stmt_fallback(
        self, frame: ForwardFrame[Address], stmt: ir.Statement
    ) -> tuple[Address, ...] | interp.SpecialValue[Address]:
        return tuple(
            (
                self.lattice.top()
                if result.type.is_subseteq(QubitType)
                else self.lattice.bottom()
            )
            for result in stmt.results
        )

    def run_method(self, method: ir.Method, args: tuple[Address, ...]):
        # NOTE: we do not support dynamic calls here, thus no need to propagate method object
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)



from typing import List

from kirin import ir, types
from bloqade import qasm2
from bloqade.noise import native
from bloqade.qbraid import schema, lowering
from kirin.dialects import func


def as_int(value: int):
    return qasm2.expr.ConstInt(value=value)


def as_float(value: float):
    return qasm2.expr.ConstFloat(value=value)


def run_assert(noise_model: schema.NoiseModel, expected_stmts: List[ir.Statement]):

    block = ir.Block(stmts=expected_stmts)
    block.args.append_from(types.MethodType, name="test_self")
    region = ir.Region(block)
    expected_func_stmt = func.Function(
        sym_name="test",
        signature=func.Signature(inputs=(), output=qasm2.types.QRegType),
        body=region,
    )

    expected_mt = ir.Method(
        mod=None,
        py_func=None,
        dialects=lowering.qbraid_noise,
        sym_name="test",
        arg_names=[],
        code=expected_func_stmt,
    )

    assert lowering.qbraid_noise.run_pass
    lowering.qbraid_noise.run_pass(expected_mt)

    mt = noise_model.lower_noise_model("test")
    try:
        assert expected_mt.code.is_structurally_equal(mt.code)
    except AssertionError as e:
        print("Generated:")
        mt.print()
        print("Expected:")
        expected_mt.print()
        raise e


def test_lowering_cz():

    survival_prob_0_value = 0.9
    survival_prob_1_value = 0.9
    survival_prob_5_value = 1.0
    survival_prob_3_value = 0.5

    px_stor_5_value = 0.1
    py_stor_5_value = 0.2
    pz_stor_5_value = 0.3

    px_sing_0_value = 0.6
    py_sing_0_value = 0.1
    pz_sing_0_value = 0.3

    px_sing_1_value = 0.3
    py_sing_1_value = 0.3
    pz_sing_1_value = 0.2

    px_sing_3_value = 0.6
    py_sing_3_value = 0.1
    pz_sing_3_value = 0.2

    px_ent_0_value = 0.0
    py_ent_0_value = 0.7
    pz_ent_0_value = 0.2

    px_ent_1_value = 0.2
    py_ent_1_value = 0.2
    pz_ent_1_value = 0.3

    operation = schema.CZ(participants=((0, 1), (3,)))
    error = schema.CZError(
        survival_prob=(
            survival_prob_0_value,
            survival_prob_1_value,
            survival_prob_5_value,
            survival_prob_3_value,
        ),
        storage_error=schema.PauliErrorModel(
            errors=((5, (px_stor_5_value, py_stor_5_value, pz_stor_5_value)),),
        ),
        entangled_error=schema.PauliErrorModel(
            errors=(
                (0, (px_ent_0_value, py_ent_0_value, pz_ent_0_value)),
                (1, (px_ent_1_value, py_ent_1_value, pz_ent_1_value)),
            ),
        ),
        single_error=schema.PauliErrorModel(
            errors=(
                (0, (px_sing_0_value, py_sing_0_value, pz_sing_0_value)),
                (1, (px_sing_1_value, py_sing_1_value, pz_sing_1_value)),
                (3, (px_sing_3_value, py_sing_3_value, pz_sing_3_value)),
            ),
        ),
    )
    gate_event = schema.GateEvent(
        operation=operation,
        error=error,
    )

    noise_model = schema.NoiseModel(
        all_qubits=(0, 1, 5, 3),
        gate_events=[gate_event],
    )

    expected: List[ir.Statement] = [
        (n_qubits := as_int(4)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (creg := qasm2.core.CRegNew(n_bits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx0.result)),
        (idx1 := as_int(1)),
        (q1 := qasm2.core.QRegGet(reg.result, idx=idx1.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx1.result)),
        (idx2 := as_int(2)),
        (q5 := qasm2.core.QRegGet(reg.result, idx=idx2.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx2.result)),
        (idx3 := as_int(3)),
        (q3 := qasm2.core.QRegGet(reg.result, idx=idx3.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx3.result)),
        (survival_prob_0 := as_float(survival_prob_0_value)),
        native.AtomLossChannel(survival_prob_0.result, q0.result),
        (survival_prob_1 := as_float(survival_prob_1_value)),
        native.AtomLossChannel(survival_prob_1.result, q1.result),
        (survival_prob_5 := as_float(survival_prob_5_value)),
        native.AtomLossChannel(survival_prob_5.result, q5.result),
        (survival_prob_3 := as_float(survival_prob_3_value)),
        native.AtomLossChannel(survival_prob_3.result, q3.result),
        (px_stor_5 := as_float(px_stor_5_value)),
        (py_stor_5 := as_float(py_stor_5_value)),
        (pz_stor_5 := as_float(pz_stor_5_value)),
        native.PauliChannel(
            px_stor_5.result, py_stor_5.result, pz_stor_5.result, q5.result
        ),
        (px_sing_0 := as_float(px_sing_0_value)),
        (py_sing_0 := as_float(py_sing_0_value)),
        (pz_sing_0 := as_float(pz_sing_0_value)),
        (px_sing_1 := as_float(px_sing_1_value)),
        (py_sing_1 := as_float(py_sing_1_value)),
        (pz_sing_1 := as_float(pz_sing_1_value)),
        native.CZPauliChannel(
            px_sing_0.result,
            py_sing_0.result,
            pz_sing_0.result,
            px_sing_1.result,
            py_sing_1.result,
            pz_sing_1.result,
            q0.result,
            q1.result,
            paired=False,
        ),
        (px_ent_0 := as_float(px_ent_0_value)),
        (py_ent_0 := as_float(py_ent_0_value)),
        (pz_ent_0 := as_float(pz_ent_0_value)),
        (px_ent_1 := as_float(px_ent_1_value)),
        (py_ent_1 := as_float(py_ent_1_value)),
        (pz_ent_1 := as_float(pz_ent_1_value)),
        native.CZPauliChannel(
            px_ent_0.result,
            py_ent_0.result,
            pz_ent_0.result,
            px_ent_1.result,
            py_ent_1.result,
            pz_ent_1.result,
            q0.result,
            q1.result,
            paired=True,
        ),
        (px_sing_3 := as_float(px_sing_3_value)),
        (py_sing_3 := as_float(py_sing_3_value)),
        (pz_sing_3 := as_float(pz_sing_3_value)),
        native.PauliChannel(
            px_sing_3.result, py_sing_3.result, pz_sing_3.result, q3.result
        ),
        qasm2.uop.CZ(q0.result, q1.result),
        func.Return(creg.result),
    ]

    run_assert(noise_model, expected)


def test_lowering_global_w():
    theta_val = 0.4
    phi_val = 0.7

    operation = schema.GlobalW(theta=theta_val, phi=phi_val)
    error = schema.SingleQubitError(
        survival_prob=(0.9,),
        operator_error=schema.PauliErrorModel(
            errors=((0, (0.1, 0.2, 0.3)),),
        ),
    )

    gate_event = schema.GateEvent(
        operation=operation,
        error=error,
    )

    noise_model = schema.NoiseModel(
        all_qubits=(0,),
        gate_events=[gate_event],
    )

    expected: List[ir.Statement] = [
        (n_qubits := as_int(1)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (creg := qasm2.core.CRegNew(n_bits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx0.result)),
        (survival_prob_0 := as_float(0.9)),
        native.AtomLossChannel(survival_prob_0.result, q0.result),
        (px_sing_0 := as_float(0.1)),
        (py_sing_0 := as_float(0.2)),
        (pz_sing_0 := as_float(0.3)),
        native.PauliChannel(
            px_sing_0.result, py_sing_0.result, pz_sing_0.result, q0.result
        ),
        (pi_theta := qasm2.expr.ConstPI()),
        (theta_num := as_float(2 * theta_val)),
        (theta := qasm2.expr.Mul(pi_theta.result, theta_num.result)),
        (pi_phi := qasm2.expr.ConstPI()),
        (phi_num := as_float(2 * (phi_val + 0.5))),
        (phi := qasm2.expr.Mul(pi_phi.result, phi_num.result)),
        (pi_lam := qasm2.expr.ConstPI()),
        (lam_num := as_float(2 * -(0.5 + phi_val))),
        (lam := qasm2.expr.Mul(pi_lam.result, lam_num.result)),
        qasm2.uop.UGate(
            theta=theta.result, phi=phi.result, lam=lam.result, qarg=q0.result
        ),
        func.Return(creg.result),
    ]

    run_assert(noise_model, expected)


def test_lowering_local_w():
    theta_val = 0.4
    phi_val = 0.7

    operation = schema.LocalW(theta=theta_val, phi=phi_val, participants=(1,))
    error = schema.SingleQubitError(
        survival_prob=(0.9, 0.4),
        operator_error=schema.PauliErrorModel(
            errors=(
                (0, (0.1, 0.2, 0.3)),
                (1, (0.5, 0.1, 0.2)),
            ),
        ),
    )

    gate_event = schema.GateEvent(
        operation=operation,
        error=error,
    )

    noise_model = schema.NoiseModel(
        all_qubits=(0, 1),
        gate_events=[gate_event],
    )

    expected: List[ir.Statement] = [
        (n_qubits := as_int(2)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (creg := qasm2.core.CRegNew(n_bits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx0.result)),
        (idx1 := as_int(1)),
        (q1 := qasm2.core.QRegGet(reg.result, idx=idx1.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx1.result)),
        (survival_prob_0 := as_float(0.9)),
        native.AtomLossChannel(survival_prob_0.result, q0.result),
        (survival_prob_1 := as_float(0.4)),
        native.AtomLossChannel(survival_prob_1.result, q1.result),
        (px_sing_0 := as_float(0.1)),
        (py_sing_0 := as_float(0.2)),
        (pz_sing_0 := as_float(0.3)),
        native.PauliChannel(
            px_sing_0.result, py_sing_0.result, pz_sing_0.result, q0.result
        ),
        (px_sing_1 := as_float(0.5)),
        (py_sing_1 := as_float(0.1)),
        (pz_sing_1 := as_float(0.2)),
        native.PauliChannel(
            px_sing_1.result, py_sing_1.result, pz_sing_1.result, q1.result
        ),
        (pi_theta := qasm2.expr.ConstPI()),
        (theta_num := as_float(2 * theta_val)),
        (theta := qasm2.expr.Mul(pi_theta.result, theta_num.result)),
        (pi_phi := qasm2.expr.ConstPI()),
        (phi_num := as_float(2 * (phi_val + 0.5))),
        (phi := qasm2.expr.Mul(pi_phi.result, phi_num.result)),
        (pi_lam := qasm2.expr.ConstPI()),
        (lam_num := as_float(2 * -(0.5 + phi_val))),
        (lam := qasm2.expr.Mul(pi_lam.result, lam_num.result)),
        qasm2.uop.UGate(
            theta=theta.result, phi=phi.result, lam=lam.result, qarg=q1.result
        ),
        func.Return(creg.result),
    ]

    run_assert(noise_model, expected)


def test_lowering_global_rz():
    phi_val = 0.7

    operation = schema.GlobalRz(phi=phi_val)
    error = schema.SingleQubitError(
        survival_prob=(0.9,),
        operator_error=schema.PauliErrorModel(
            errors=((0, (0.1, 0.2, 0.3)),),
        ),
    )

    gate_event = schema.GateEvent(
        operation=operation,
        error=error,
    )

    noise_model = schema.NoiseModel(
        all_qubits=(0,),
        gate_events=[gate_event],
    )

    expected: List[ir.Statement] = [
        (n_qubits := as_int(1)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (creg := qasm2.core.CRegNew(n_bits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx0.result)),
        (survival_prob_0 := as_float(0.9)),
        native.AtomLossChannel(survival_prob_0.result, q0.result),
        (px_sing_0 := as_float(0.1)),
        (py_sing_0 := as_float(0.2)),
        (pz_sing_0 := as_float(0.3)),
        native.PauliChannel(
            px_sing_0.result, py_sing_0.result, pz_sing_0.result, q0.result
        ),
        (theta_pi := qasm2.expr.ConstPI()),
        (theta_num := as_float(2 * phi_val)),
        (theta := qasm2.expr.Mul(theta_pi.result, theta_num.result)),
        qasm2.uop.RZ(theta=theta.result, qarg=q0.result),
        func.Return(creg.result),
    ]

    run_assert(noise_model, expected)


def test_lowering_local_rz():
    phi_val = 0.7

    operation = schema.LocalRz(phi=phi_val, participants=(1,))
    error = schema.SingleQubitError(
        survival_prob=(0.9, 0.4),
        operator_error=schema.PauliErrorModel(
            errors=(
                (0, (0.1, 0.2, 0.3)),
                (1, (0.5, 0.1, 0.2)),
            ),
        ),
    )

    gate_event = schema.GateEvent(
        operation=operation,
        error=error,
    )

    noise_model = schema.NoiseModel(
        all_qubits=(0, 1),
        gate_events=[gate_event],
    )

    expected: List[ir.Statement] = [
        (n_qubits := as_int(2)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (creg := qasm2.core.CRegNew(n_bits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx0.result)),
        (idx1 := as_int(1)),
        (q1 := qasm2.core.QRegGet(reg.result, idx=idx1.result)),
        (_ := qasm2.core.CRegGet(creg.result, idx=idx1.result)),
        (survival_prob_0 := as_float(0.9)),
        native.AtomLossChannel(survival_prob_0.result, q0.result),
        (survival_prob_1 := as_float(0.4)),
        native.AtomLossChannel(survival_prob_1.result, q1.result),
        (px_sing_0 := as_float(0.1)),
        (py_sing_0 := as_float(0.2)),
        (pz_sing_0 := as_float(0.3)),
        native.PauliChannel(
            px_sing_0.result, py_sing_0.result, pz_sing_0.result, q0.result
        ),
        (px_sing_1 := as_float(0.5)),
        (py_sing_1 := as_float(0.1)),
        (pz_sing_1 := as_float(0.2)),
        native.PauliChannel(
            px_sing_1.result, py_sing_1.result, pz_sing_1.result, q1.result
        ),
        (theta_pi := qasm2.expr.ConstPI()),
        (theta_num := as_float(2 * phi_val)),
        (theta := qasm2.expr.Mul(theta_pi.result, theta_num.result)),
        qasm2.uop.RZ(theta=theta.result, qarg=q1.result),
        func.Return(creg.result),
    ]

    run_assert(noise_model, expected)



from . import aux as aux, gate as gate, noise as noise, collapse as collapse
from .aux.stmts import *  # noqa F403
from .gate.stmts import *  # noqa F403
from .noise.stmts import *  # noqa F403
from .collapse.stmts import *  # noqa F403



import ast

from kirin.lowering import Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError

from . import stmts
from ._dialect import dialect


@dialect.register
class StimAuxLowering(FromPythonAST):

    def _const_stmt(
        self, state: LoweringState, value: int | float | str | bool
    ) -> stmts.ConstInt | stmts.ConstFloat | stmts.ConstStr | stmts.ConstBool:

        if isinstance(value, bool):
            return stmts.ConstBool(value=value)
        elif isinstance(value, int):
            return stmts.ConstInt(value=value)
        elif isinstance(value, float):
            return stmts.ConstFloat(value=value)
        elif isinstance(value, str):
            return stmts.ConstStr(value=value)
        else:
            raise DialectLoweringError(f"unsupported Stim constant type {type(value)}")

    def lower_Constant(self, state: LoweringState, node: ast.Constant) -> Result:
        stmt = self._const_stmt(state, node.value)
        return Result(state.append_stmt(stmt))

    def lower_Expr(self, state: LoweringState, node: ast.Expr) -> Result:
        return state.visit(node.value)

    def lower_UnaryOp(self, state: LoweringState, node: ast.UnaryOp) -> Result:
        if isinstance(node.op, ast.USub):
            value = state.visit(node.operand).expect_one()
            stmt = stmts.Neg(operand=value)
            return Result(state.append_stmt(stmt))
        else:
            raise DialectLoweringError(f"unsupported Stim unaryop {node.op}")



from typing import Generic, TypeVar

from . import ast

T = TypeVar("T")


class Visitor(Generic[T]):

    def visit(self, node: ast.Node) -> T:
        method_name = f"visit_{node.__class__.__name__}"
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node: ast.Node) -> T:
        raise NotImplementedError(f"No visit_{node.__class__.__name__} method")



from .stmts import (
    PauliChannel as PauliChannel,
    CZPauliChannel as CZPauliChannel,
    AtomLossChannel as AtomLossChannel,
)
from .rewrite import RemoveNoisePass as RemoveNoisePass
from ._dialect import dialect as dialect
from ._wrappers import (
    pauli_channel as pauli_channel,
    cz_pauli_channel as cz_pauli_channel,
    atom_loss_channel as atom_loss_channel,
)



from bloqade import qasm2


@qasm2.gate
def custom_gate(a: qasm2.Qubit, b: qasm2.Qubit):
    qasm2.cx(a, b)


@qasm2.main
def main():
    qreg = qasm2.qreg(4)
    creg = qasm2.creg(2)
    qasm2.cx(qreg[0], qreg[1])
    qasm2.reset(qreg[0])
    # qasm2.parallel.cz(ctrls=[qreg[0], qreg[1]], qargs=[qreg[2], qreg[3]])
    qasm2.measure(qreg[0], creg[0])
    if creg[0] == 1:
        qasm2.reset(qreg[1])
    custom_gate(qreg[0], qreg[1])


main.print()
custom_gate.print()

target = qasm2.emit.QASM2(custom_gate=True)
ast = target.emit(main)
qasm2.parse.pprint(ast)



import dataclasses
from typing import Any, Dict


@dataclasses.dataclass(frozen=True)
class GateEvent:
    cls_name: str
    kwargs: Dict[str, Any]
    duration: float

    def to_json(self) -> Dict[str, Any]:
        return {
            "cls_name": self.cls_name,
            "kwargs": self.kwargs,
            "duration": self.duration,
        }

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "GateEvent":
        return cls(
            json_dict["cls_name"],
            json_dict["kwargs"],
            json_dict["duration"],
        )



"""
Passes that deal with global gates. As of now, only one rewrite pass exists
which converts global gates to single qubit gates.
"""

from kirin import ir
from kirin.rewrite import cse, dce, walk, result
from bloqade.analysis import address
from kirin.passes.abc import Pass
from bloqade.qasm2.rewrite import GlobalToUOpRule


class GlobalToUOP(Pass):
    """Pass to convert Global gates into single gates.

    This pass rewrites the global unitary gate from the `qasm2.glob` dialect into multiple
    single gates in the `qasm2.uop` dialect, bringing the program closer to
    conforming to standard QASM2 syntax.


    ## Usage Examples
    ```
    # Define kernel
    @qasm2.extended
    def main():
        q1 = qasm2.qreg(1)
        q2 = qasm2.qreg(2)

        theta = 1.3
        phi = 1.1
        lam = 1.2

        qasm2.glob.u(theta=theta, phi=phi, lam=lam, registers=[q1, q2])

    GlobalToUOP(dialects=main.dialects)(main)

    # Run rewrite
    GlobalToUOP(main.dialects)(main)
    ```

    The `qasm2.glob.u` statement has been rewritten to individual gates:

    ```
    qasm2.uop.u(q1[0], theta, phi, lam)
    qasm2.uop.u(q1[0], theta, phi, lam)
    qasm2.uop.u(q2[1], theta, phi, lam)
    ```
    """

    def generate_rule(self, mt: ir.Method) -> GlobalToUOpRule:
        results, _ = address.AddressAnalysis(mt.dialects).run_analysis(mt)

        # You can't hash the address register because it contains a Sequence type,
        # need to juggle things around in lists
        encountered_addr_regs = []
        encountered_addr_reg_ssas = []

        in_global_addr_regs = []
        in_global_addr_reg_ssas = []

        for ssa, addr in results.entries.items():

            # Find all the times an AddressReg is encountered
            if isinstance(addr, address.AddressReg):
                encountered_addr_regs.append(addr)
                encountered_addr_reg_ssas.append(ssa)

            # When we encounter an AddressTuple,
            # check which registers are referenced by it and
            # after verifying, use the SSA values and the registers encountered
            # previously to construct to the GlobalToUOpRule
            if isinstance(addr, address.AddressTuple):
                for encountered_addr_reg, encountered_addr_reg_ssa in zip(
                    encountered_addr_regs, encountered_addr_reg_ssas
                ):
                    if encountered_addr_reg in addr.data:
                        in_global_addr_regs.append(encountered_addr_reg)
                        in_global_addr_reg_ssas.append(encountered_addr_reg_ssa)

        return GlobalToUOpRule(
            address_regs=in_global_addr_regs, address_reg_ssas=in_global_addr_reg_ssas
        )

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        rewriter = walk.Walk(self.generate_rule(mt))
        result = rewriter.rewrite(mt.code)

        result = walk.Walk(dce.DeadCodeElimination()).rewrite(mt.code)
        result = walk.Walk(cse.CommonSubexpressionElimination()).rewrite(mt.code)

        return result



# %% [markdown]
# # Quantum Fourier Transform
# In this example, we will explore the Quantum Fourier Transform (QFT) circuit using
# recursion and iteration -- a convenient way to implement the QFT circuit using
# our high-level programming features.
#
# To begin, we will import the `qasm2` module from the `bloqade` package and the `PyQrack`
# backend from the `bloqade.pyqrack` module, which can be installed via
#
# ```bash
# pip install bloqade[pyqrack]
# ```
# or
# ```bash
# pip install bloqade[pyqrack-cpu]
# ```
# depending on the hardware and OS you have.
# %%
import math

from bloqade import qasm2
from bloqade.pyqrack import PyQrack

# In the following, we will define the Quantum Fourier Transform (QFT) circuit using recursion
# inside a kernel function `qft`. The `qft` function takes two arguments: a quantum register `qreg`
# and an integer `n` representing the number of qubits we want to apply the QFT circuit to.
# %%
pi = math.pi


@qasm2.extended
def qft(qreg: qasm2.QReg, n: int):
    if n == 0:
        return qreg

    qasm2.h(qreg[0])
    for i in range(1, n):
        qasm2.cu1(qreg[i], qreg[0], 2 * pi / 2**i)
    qft(qreg, n - 1)
    return qreg


# %% [markdown]
# Next, we will call this kernel function `qft` inside a `main` function to check if
# the QFT circuit is correctly implemented. We will use a quantum register of size 3.


# %%
@qasm2.extended
def main():
    return qft(qasm2.qreg(3), 3)


# %% [markdown]
# Finally, we will run the `main` function on the `PyQrack` backend and print the quantum register
# to see the final state of the qubits after applying the QFT circuit.

# %%
device = PyQrack()
qreg = device.run(main)
print(qreg)

# %% [markdown]
# we can also emit the QASM2 code for the `main` function and print it to see the QASM2 code
# that corresponds to the QFT circuit.

# %%
from bloqade.qasm2.emit import QASM2  # noqa: E402
from bloqade.qasm2.parse import pprint  # noqa: E402

target = QASM2()
ast = target.emit(main)
pprint(ast)



from .pp import SPP as SPP
from .control_2q import CX as CX, CY as CY, CZ as CZ
from .clifford_1q import (
    H as H,
    S as S,
    X as X,
    Y as Y,
    Z as Z,
    SqrtX as SqrtX,
    SqrtY as SqrtY,
    SqrtZ as SqrtZ,
    Identity as Identity,
)
from .clifford_2q import Swap as Swap



# %%
import math
from typing import Any

# %%
import networkx as nx
from bloqade import qasm2
from kirin.dialects import py, ilist

# %%
# Define the problem instance as a MaxCut problem on a graph
pi = math.pi
N = 64
G = nx.random_regular_graph(3, N, seed=42)


# %%
def qaoa_sequential(G):

    edges = list(G.edges)
    nodes = list(G.nodes)

    @qasm2.extended
    def kernel(gamma: ilist.IList[float, Any], delta: ilist.IList[float, Any]):
        qreg = qasm2.qreg(len(nodes))
        for i in range(N):
            qasm2.h(qreg[i])

        for i in range(len(gamma)):
            for j in range(len(edges)):
                edge = edges[j]
                qasm2.cx(qreg[edge[0]], qreg[edge[1]])
                qasm2.rz(qreg[edge[1]], gamma[i])
                qasm2.cx(qreg[edge[0]], qreg[edge[1]])

            for j in range(N):
                qasm2.rx(qreg[j], delta[i])

        return qreg

    return kernel


# %%
def qaoa_simd(G):

    left_ids = ilist.IList([edge[0] for edge in G.edges])
    right_ids = ilist.IList([edge[1] for edge in G.edges])
    nodes = list(G.nodes)

    @qasm2.extended
    def parallel_h(qargs: ilist.IList[qasm2.Qubit, Any]):
        qasm2.parallel.u(qargs=qargs, theta=pi / 2, phi=0.0, lam=pi)

    @qasm2.extended
    def parallel_cx(
        ctrls: ilist.IList[qasm2.Qubit, Any], qargs: ilist.IList[qasm2.Qubit, Any]
    ):
        parallel_h(qargs)
        qasm2.parallel.cz(ctrls, qargs)
        parallel_h(qargs)

    @qasm2.extended
    def parallel_cz_phase(
        ctrls: ilist.IList[qasm2.Qubit, Any],
        qargs: ilist.IList[qasm2.Qubit, Any],
        gamma: float,
    ):
        parallel_cx(ctrls, qargs)
        qasm2.parallel.rz(qargs, gamma)
        parallel_cx(ctrls, qargs)

    @qasm2.extended
    def kernel(gamma: ilist.IList[float, Any], beta: ilist.IList[float, Any]):
        qreg = qasm2.qreg(len(nodes))

        def get_qubit(x: int):
            return qreg[x]

        ctrls = ilist.map(fn=get_qubit, collection=left_ids)
        qargs = ilist.map(fn=get_qubit, collection=right_ids)
        all_qubits = ilist.map(fn=get_qubit, collection=range(N))

        parallel_h(all_qubits)
        for i in range(len(gamma)):
            parallel_cz_phase(ctrls, qargs, gamma[i])
            qasm2.parallel.u(all_qubits, beta[i], 0.0, 0.0)

        return qreg

    return kernel


# %%
print("--- Sequential ---")
qaoa_sequential(G).code.print()

# %%
kernel = qaoa_simd(G)

print("\n\n--- Simd ---")
kernel.print()


# %%
@qasm2.extended
def main():
    kernel([0.1, 0.2], [0.3, 0.4])


# %%
target = qasm2.emit.QASM2(
    main_target=qasm2.main.union([qasm2.dialects.parallel, ilist, py.constant])
)
ast = target.emit(main)
qasm2.parse.pprint(ast)



from kirin import ir, types
from kirin.decl import info, statement
from kirin.print.printer import Printer
from kirin.dialects.func.attrs import Signature

from ._dialect import dialect


class GateFuncOpCallableInterface(ir.CallableStmtInterface["GateFunction"]):

    @classmethod
    def get_callable_region(cls, stmt: "GateFunction") -> ir.Region:
        return stmt.body


@statement(dialect=dialect)
class GateFunction(ir.Statement):
    """Special Function for qasm2 gate subroutine."""

    name = "gate.func"
    traits = frozenset(
        {
            ir.IsolatedFromAbove(),
            ir.SymbolOpInterface(),
            ir.HasSignature(),
            GateFuncOpCallableInterface(),
        }
    )
    sym_name: str = info.attribute()
    signature: Signature = info.attribute()
    body: ir.Region = info.region(multi=True)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="red"):
            printer.plain_print(self.name + " ")

        with printer.rich(style="cyan"):
            printer.plain_print(self.sym_name)

        self.signature.print_impl(printer)
        printer.plain_print(" ")
        self.body.print_impl(printer)

        with printer.rich(style="black"):
            printer.plain_print(f" // gate.func {self.sym_name}")


@statement(dialect=dialect)
class ConstInt(ir.Statement):
    """IR Statement representing a constant integer value."""

    name = "constant.int"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: int = info.attribute(types.Int)
    """value (int): The constant integer value."""
    result: ir.ResultValue = info.result(types.Int)
    """result (Int): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class ConstFloat(ir.Statement):
    """IR Statement representing a constant float value."""

    name = "constant.float"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: float = info.attribute(types.Float)
    """value (float): The constant float value."""
    result: ir.ResultValue = info.result(types.Float)
    """result (Float): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class ConstPI(ir.Statement):
    """The constant value of PI."""

    # this is marked as constant but not pure.
    name = "constant.pi"
    traits = frozenset({ir.ConstantLike(), ir.FromPythonCall()})
    result: ir.ResultValue = info.result(types.Float)
    """result (ConstPI): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print("PI")
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


# QASM 2.0 arithmetic operations
PyNum = types.Union(types.Int, types.Float)


@statement(dialect=dialect)
class Neg(ir.Statement):
    """Negate a number."""

    name = "neg"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to negate."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The negated number."""


@statement(dialect=dialect)
class Sin(ir.Statement):
    """Take the sine of a number."""

    name = "sin"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the sine of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The sine of the number."""


@statement(dialect=dialect)
class Cos(ir.Statement):
    """Take the cosine of a number."""

    name = "cos"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the cosine of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The cosine of the number."""


@statement(dialect=dialect)
class Tan(ir.Statement):
    """Take the tangent of a number."""

    name = "tan"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the tangent of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The tangent of the number."""


@statement(dialect=dialect)
class Exp(ir.Statement):
    """Take the exponential of a number."""

    name = "exp"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the exponential of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The exponential of the number."""


@statement(dialect=dialect)
class Log(ir.Statement):
    """Take the natural log of a number."""

    name = "ln"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the natural log of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The natural log of the number."""


@statement(dialect=dialect)
class Sqrt(ir.Statement):
    """Take the square root of a number."""

    name = "sqrt"
    traits = frozenset({ir.FromPythonCall()})
    value: ir.SSAValue = info.argument(PyNum)
    """value (Union[int, float]): The number to take the square root of."""
    result: ir.ResultValue = info.result(PyNum)
    """result (float): The square root of the number."""


@statement(dialect=dialect)
class Add(ir.Statement):
    """Add two numbers."""

    name = "add"
    traits = frozenset({ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(PyNum)
    """lhs (Union[int, float]): The left-hand side of the addition."""
    rhs: ir.SSAValue = info.argument(PyNum)
    """rhs (Union[int, float]): The right-hand side of the addition."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The result of the addition."""


@statement(dialect=dialect)
class Sub(ir.Statement):
    """Subtract two numbers."""

    name = "sub"
    traits = frozenset({ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(PyNum)
    """lhs (Union[int, float]): The left-hand side of the subtraction."""
    rhs: ir.SSAValue = info.argument(PyNum)
    """rhs (Union[int, float]): The right-hand side of the subtraction."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The result of the subtraction."""


@statement(dialect=dialect)
class Mul(ir.Statement):
    """Multiply two numbers."""

    name = "mul"
    traits = frozenset({ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(PyNum)
    """lhs (Union[int, float]): The left-hand side of the multiplication."""
    rhs: ir.SSAValue = info.argument(PyNum)
    """rhs (Union[int, float]): The right-hand side of the multiplication."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The result of the multiplication."""


@statement(dialect=dialect)
class Pow(ir.Statement):
    """Take the power of a number."""

    name = "pow"
    traits = frozenset({ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(PyNum)
    """lhs (Union[int, float]): The base."""
    rhs: ir.SSAValue = info.argument(PyNum)
    """rhs (Union[int, float]): The exponent."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The result of the power operation."""


@statement(dialect=dialect)
class Div(ir.Statement):
    """Divide two numbers."""

    name = "div"
    traits = frozenset({ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(PyNum)
    """lhs (Union[int, float]): The numerator."""
    rhs: ir.SSAValue = info.argument(PyNum)
    """rhs (Union[int, float]): The denominator."""
    result: ir.ResultValue = info.result(PyNum)
    """result (Union[int, float]): The result of the division."""



from typing import Sequence, final
from dataclasses import dataclass

from kirin.lattice import (
    SingletonMeta,
    BoundedLattice,
    SimpleJoinMixin,
    SimpleMeetMixin,
)


@dataclass
class Address(
    SimpleJoinMixin["Address"],
    SimpleMeetMixin["Address"],
    BoundedLattice["Address"],
):

    @classmethod
    def bottom(cls) -> "Address":
        return NotQubit()

    @classmethod
    def top(cls) -> "Address":
        return AnyAddress()


@final
@dataclass
class NotQubit(Address, metaclass=SingletonMeta):

    def is_subseteq(self, other: Address) -> bool:
        return True


@final
@dataclass
class AnyAddress(Address, metaclass=SingletonMeta):

    def is_subseteq(self, other: Address) -> bool:
        return isinstance(other, AnyAddress)


@final
@dataclass
class AddressTuple(Address):
    data: tuple[Address, ...]

    def is_subseteq(self, other: Address) -> bool:
        if isinstance(other, AddressTuple):
            return all(a.is_subseteq(b) for a, b in zip(self.data, other.data))
        return False


@final
@dataclass
class AddressReg(Address):
    data: Sequence[int]

    def is_subseteq(self, other: Address) -> bool:
        if isinstance(other, AddressReg):
            return self.data == other.data
        return False


@final
@dataclass
class AddressQubit(Address):
    data: int

    def is_subseteq(self, other: Address) -> bool:
        if isinstance(other, AddressQubit):
            return self.data == other.data
        return False



from dataclasses import dataclass

from kirin import types


@dataclass
class RecordResult:
    value: int


@dataclass
class PauliString:
    string: tuple[str, ...]
    flipped: tuple[bool, ...]
    targets: tuple[int, ...]


RecordType = types.PyClass(RecordResult)
PauliStringType = types.PyClass(PauliString)



# %% [markdown]
# # Pauli Exponentiation for Quantum Simulation
# In this example, we will consider a simple Pauli Exponentiation circuit.

# %%
import math

from bloqade import qasm2

# %% [markdown]
# First, we define the `zzzz_gadget` function which is a simple implementation of Pauli Z exponentiation
# with a parameterized angle `gamma`.


# %%
@qasm2.extended
def zzzz_gadget(targets: tuple[qasm2.Qubit, ...], gamma: float):
    for i in range(len(targets) - 1):
        qasm2.cx(targets[i], targets[i + 1])

    qasm2.rz(targets[-1], gamma)

    for j in range(len(targets) - 1):
        qasm2.cx(targets[-j - 1], targets[-j - 2])


# %% [markdown]
# Next, we define the `pauli_basis_change` function which is a simple implementation of Pauli basis change
# with a parameterized start and end Pauli basis.


# %%
@qasm2.extended
def pauli_basis_change(targets: tuple[qasm2.Qubit, ...], start: str, end: str):
    # assert len(targets) == len(start)
    # assert len(targets) == len(end)

    # for qubit, start_pauli, end_pauli in zip(targets, start, end):
    for i in range(len(targets)):
        qubit = targets[i]
        start_pauli = start[i]
        end_pauli = end[i]

        target = start_pauli + end_pauli
        if target == "ZX":
            qasm2.ry(qubit, math.pi / 2)
        elif target == "ZY":
            qasm2.rx(qubit, -math.pi / 2)
        # elif target == "ZZ":
        #     pass
        # elif target == "XX":
        #     pass
        elif target == "XY":
            qasm2.rz(qubit, math.pi / 2)
        elif target == "XZ":
            qasm2.ry(qubit, -math.pi / 2)
        elif target == "YX":
            qasm2.rz(qubit, -math.pi / 2)
        # elif target == "YY":
        #     pass
        elif target == "YZ":
            qasm2.rx(qubit, math.pi / 2)


# %% [markdown]
# Putting it all together, we define the `pauli_exponential` function which is a simple implementation of Pauli Exponentiation
# with a parameterized Pauli basis and angle `gamma`.
# %%
@qasm2.extended
def pauli_exponential(targets: tuple[qasm2.Qubit, ...], pauli: str, gamma: float):
    # assert len(targets) == len(pauli)

    pauli_basis_change(targets=targets, start="Z" * len(targets), end=pauli)
    zzzz_gadget(targets=targets, gamma=gamma)
    pauli_basis_change(targets=targets, start=pauli, end="Z" * len(targets))


# %% [markdown]
# Finally, we define the `main` function as the entry point of the program.


# %%
@qasm2.extended
def main():
    register = qasm2.qreg(4)
    pauli_exponential((register[0], register[1], register[2]), "ZXY", 0.5)


# %% [markdown]
# we can now ask the compiler to emit the QASM2 code for the `main` function.
# %%
target = qasm2.emit.QASM2()
ast = target.emit(main)
qasm2.parse.pprint(ast)



from typing import List, Tuple, Union, Generic, Literal, TypeVar

from pydantic import Field, BaseModel


class Operation(BaseModel, frozen=True, extra="forbid"):
    """Base class for operations."""

    op_type: str = Field(init=False)


class CZ(Operation):
    """A CZ gate operation.

    Fields:
        op_type (str): The type of operation (Literal["CZ"]).
        participants (Tuple[Union[Tuple[int], Tuple[int, int]], ...]): The qubit indices that are participating in the CZ gate.

    """

    op_type: Literal["CZ"] = Field(init=False, default="CZ")
    participants: Tuple[Union[Tuple[int], Tuple[int, int]], ...]


class GlobalRz(Operation):
    """GlobalRz operation.

    Fields:
        op_type (str): The type of operation (Literal["GlobalRz"]).
        phi (float): The angle of rotation.
    """

    op_type: Literal["GlobalRz"] = Field(init=False, default="GlobalRz")
    phi: float


class GlobalW(Operation):
    """GlobalW operation.

    Fields:
        op_type (str): The type of operation (Literal["GlobalW"]).
        theta (float): The angle of rotation.
        phi (float): The angle of rotation.
    """

    op_type: Literal["GlobalW"] = Field(init=False, default="GlobalW")
    theta: float
    phi: float


class LocalRz(Operation):
    """LocalRz operation.

    Fields:
        op_type (str): The type of operation (Literal["LocalRz"]).
        participants (Tuple[int, ...]): The qubit indices that are participating in the local Rz gate.
        phi (float): The angle of rotation.

    """

    op_type: Literal["LocalRz"] = Field(init=False, default="LocalRz")
    participants: Tuple[int, ...]
    phi: float


class LocalW(Operation):
    """LocalW operation.

    Fields:
        op_type (str): The type of operation (Literal["LocalW"]).
        participants (Tuple[int, ...]): The qubit indices that are participating in the local W gate.
        theta (float): The angle of rotation.
        phi (float): The angle of rotation.

    """

    op_type: Literal["LocalW"] = Field(init=False, default="LocalW")
    participants: Tuple[int, ...]
    theta: float
    phi: float


class Measurement(Operation):
    """Measurement operation.

    Fields:
        op_type (str): The type of operation (Literal["Measurement"]).
        measure_tag (str): The tag to use for the measurement.
        participants (Tuple[int, ...]): The qubit indices that are participating in the measurement.

    """

    op_type: Literal["Measurement"] = Field(init=False, default="Measurement")
    measure_tag: str = Field(default="m")
    participants: Tuple[int, ...]


OperationType = TypeVar(
    "OperationType", bound=Union[CZ, GlobalRz, GlobalW, LocalRz, LocalW, Measurement]
)


class ErrorModel(BaseModel, frozen=True, extra="forbid"):
    """Base class for error models."""

    error_model_type: str = Field(init=False)


class PauliErrorModel(ErrorModel):
    """Pauli error model.

    Fields:
        error_model_type (str): The type of error model (Literal["PauliNoise"]).
        errors (Tuple[Tuple[int, Tuple[float, float, float]], ...]): The qubit indices and the error rates for each qubit.

    """

    error_model_type: Literal["PauliNoise"] = Field(default="PauliNoise", init=False)
    errors: Tuple[Tuple[int, Tuple[float, float, float]], ...] = Field(
        default_factory=tuple
    )


ErrorModelType = TypeVar("ErrorModelType", bound=PauliErrorModel)


class ErrorOperation(BaseModel, Generic[ErrorModelType], frozen=True, extra="forbid"):
    """Base class for error operations."""

    error_type: str = Field(init=False)
    survival_prob: Tuple[float, ...]


class CZError(ErrorOperation[ErrorModelType]):
    """CZError operation.

    Fields:
        survival_prob (Tuple[float, ...]): The survival probabilities for each qubit.
        error_type (str): The type of error (Literal["CZError"]).
        storage_error (ErrorModelType): The error model for storage.
        entangled_error (ErrorModelType): The error model for entangled qubits.
        single_error (ErrorModelType): The error model for single qubits.

    """

    error_type: Literal["CZError"] = Field(default="CZError", init=False)
    storage_error: ErrorModelType
    entangled_error: ErrorModelType
    single_error: ErrorModelType


class SingleQubitError(ErrorOperation[ErrorModelType]):
    """SingleQubitError operation.

    Fields:
        survival_prob (Tuple[float, ...]): The survival probabilities for each qubit.
        error_type (str): The type of error (Literal["SingleQubitError"]).
        operator_error (ErrorModelType): The error model for the single qubit.

    """

    error_type: Literal["SingleQubitError"] = Field(
        default="SingleQubitError", init=False
    )
    operator_error: ErrorModelType


class GateEvent(BaseModel, Generic[ErrorModelType], frozen=True, extra="forbid"):
    """A gate event.

    Fields:
        error (Union[SingleQubitError[ErrorModelType], CZError[ErrorModelType]]): The error model for the gate event.
        operation (OperationType): The operation for the gate event.

    """

    error: Union[SingleQubitError[ErrorModelType], CZError[ErrorModelType]] = Field(
        union_mode="left_to_right", discriminator="error_type"
    )
    operation: OperationType = Field(
        union_mode="left_to_right", discriminator="op_type"
    )

    def __pydantic_post_init__(self):
        assert (isinstance(self.operation, CZ) and isinstance(self.error, CZError)) or (
            not isinstance(self.operation, CZ)
            and isinstance(self.error, SingleQubitError)
        ), "Operation and error must be of the same type"


class NoiseModel(BaseModel, Generic[ErrorModelType], extra="forbid"):
    """Noise model for a circuit.

    Fields:
        all_qubits (Tuple[int, ...]): The qubit indices for the noise model.
        gate_events (List[GateEvent[ErrorModelType]]): The gate events for the noise model.

    """

    all_qubits: Tuple[int, ...] = Field(default_factory=tuple)
    gate_events: List[GateEvent[ErrorModelType]] = Field(default_factory=list)

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the noise model."""
        return len(self.all_qubits)

    def __add__(self, other: "NoiseModel") -> "NoiseModel":
        if not isinstance(other, NoiseModel):
            raise ValueError(f"Cannot add {type(other)} to Circuit")

        if self.all_qubits != other.all_qubits:
            raise ValueError("Circuits must have the same number of qubits")

        return NoiseModel(
            all_qubits=self.all_qubits,
            gate_events=self.gate_events + other.gate_events,
        )

    def lower_noise_model(self, sym_name: str, return_qreg: bool = False):
        """Lower the noise model to a method.

        Args:
            sym_name (str): The name of the method to generate.
            return_qreg (bool): Whether to return the quantum register after the method
                has completed execution. Useful for obtaining the full state vector.

        Returns:
            Method: The generated kirin method.

        """
        from bloqade.qbraid.lowering import Lowering

        return Lowering().lower(sym_name, self, return_qreg)

    def decompiled_circuit(self) -> str:
        """Clean the circuit of noise.

        Returns:
            str: The decompiled circuit from hardware execution.

        """
        from kirin import ir
        from bloqade.noise import native
        from bloqade.qasm2.emit import QASM2

        mt = self.lower_noise_model("method")

        native.RemoveNoisePass(mt.dialects)(mt)
        mt.dialects = ir.DialectGroup(
            mt.dialects.data.symmetric_difference([native.dialect])
        )

        return QASM2(qelib1=True).emit_str(mt)



from dataclasses import field, dataclass

from kirin import print

from .ast import (
    OPENQASM,
    Pi,
    Bit,
    Cmp,
    Call,
    CReg,
    Gate,
    Name,
    QReg,
    BinOp,
    Kirin,
    Reset,
    UGate,
    CXGate,
    IfStmt,
    Number,
    Opaque,
    Barrier,
    Comment,
    Include,
    Measure,
    UnaryOp,
    GlobUGate,
    ParaCZGate,
    ParaRZGate,
    ParaU3Gate,
    Instruction,
    MainProgram,
    NoisePAULI1,
    ParallelQArgs,
)
from .visitor import Visitor


@dataclass
class ColorScheme:
    comment: str = "bright_black"
    keyword: str = "red"
    symbol: str = "cyan"
    string: str = "yellow"
    number: str = "green"
    irrational: str = "magenta"


@dataclass
class PrintState:
    indent: int = 0
    result_width: int = 0
    rich_style: str | None = None
    rich_highlight: bool | None = False
    indent_marks: list[int] = field(default_factory=list)
    messages: list[str] = field(default_factory=list)


class Printer(print.Printer, Visitor[None]):

    def visit_MainProgram(self, node: MainProgram) -> None:
        self.print_indent()
        self.visit(node.header)
        self.print_newline()
        for stmt in node.statements:
            self.visit(stmt)
            self.print_newline()

    def visit_OPENQASM(self, node: OPENQASM) -> None:
        self.plain_print(
            f"OPENQASM {node.version.major}.{node.version.minor}",
            style="comment",
        )
        self.plain_print(";")

    def visit_Kirin(self, node: Kirin) -> None:
        self.plain_print(
            "KIRIN " + "{" + ",".join(sorted(node.dialects)) + "}", style="comment"
        )
        self.plain_print(";")

    def visit_Include(self, node: Include) -> None:
        self.plain_print("include", style="keyword")
        self.plain_print(" ")
        self.plain_print('"', node.filename, '"', style="string")
        self.plain_print(";")

    def visit_Barrier(self, node: Barrier) -> None:
        self.print_indent()
        self.plain_print("barrier", style="keyword")
        self.plain_print(" ")
        self.print_seq(node.qargs, emit=self.visit)
        self.plain_print(";")

    def visit_Instruction(self, node: Instruction) -> None:
        self.visit_Name(node.name)
        self.plain_print(" ")
        if node.params:
            self.print_seq(
                node.params, delim=", ", prefix="(", suffix=") ", emit=self.visit
            )
        self.print_seq(node.qargs, emit=self.visit)
        self.plain_print(";")

    def visit_Comment(self, node: Comment) -> None:
        self.plain_print("// ", node.text, style="comment")

    def visit_CReg(self, node: CReg) -> None:
        self.plain_print("creg", style="keyword")
        self.plain_print(f" {node.name}[{node.size}]")
        self.plain_print(";")

    def visit_QReg(self, node: QReg) -> None:
        self.plain_print("qreg", style="keyword")
        self.plain_print(f" {node.name}[{node.size}]")
        self.plain_print(";")

    def visit_CXGate(self, node: CXGate) -> None:
        self.plain_print("CX", style="keyword")
        self.plain_print(" ")
        self.visit(node.ctrl)
        self.plain_print(", ")
        self.visit(node.qarg)
        self.plain_print(";")

    def visit_UGate(self, node: UGate) -> None:
        self.plain_print("U", style="keyword")
        self.plain_print("(")
        self.visit(node.theta)
        self.plain_print(", ")
        self.visit(node.phi)
        self.plain_print(", ")
        self.visit(node.lam)
        self.plain_print(") ")
        self.visit(node.qarg)
        self.plain_print(";")

    def visit_Measure(self, node: Measure) -> None:
        self.plain_print("measure", style="keyword")
        self.plain_print(" ")
        self.visit(node.qarg)
        self.plain_print(" -> ")
        self.visit(node.carg)
        self.plain_print(";")

    def visit_Reset(self, node: Reset) -> None:
        self.plain_print("reset ")
        self.visit(node.qarg)
        self.plain_print(";")

    def visit_Opaque(self, node: Opaque) -> None:
        self.plain_print("opaque ", style="keyword")
        if node.cparams:
            self.print_seq(
                node.cparams, delim=", ", prefix="(", suffix=")", emit=self.visit
            )

        if node.qparams:
            self.plain_print(" ")
            self.print_seq(node.qparams, delim=", ", emit=self.visit)
        self.plain_print(";")

    def visit_Gate(self, node: Gate) -> None:
        self.plain_print("gate ", style="keyword")
        self.plain_print(node.name, style="symbol")
        if node.cparams:
            self.print_seq(
                node.cparams, delim=", ", prefix="(", suffix=")", emit=self.plain_print
            )

        if node.qparams:
            self.plain_print(" ")
            self.print_seq(node.qparams, delim=", ", emit=self.plain_print)

        self.plain_print(" {")
        with self.indent():
            self.print_newline()
            for idx, stmt in enumerate(node.body):
                self.visit(stmt)
                if idx < len(node.body) - 1:
                    self.print_newline()
        self.print_newline()
        self.plain_print("}")

    def visit_IfStmt(self, node: IfStmt) -> None:
        self.plain_print("if", style="keyword")
        self.visit(node.cond)
        if len(node.body) == 1:  # inline if
            self.visit(node.body[0])
        else:
            self.plain_print("{")
            with self.indent():
                self.print_newline()
                for idx, stmt in enumerate(node.body):
                    self.visit(stmt)
                    if idx < len(node.body) - 1:
                        self.print_newline()
            self.print_newline()
            self.plain_print("}")

    def visit_Cmp(self, node: Cmp) -> None:
        self.plain_print(" (")
        self.visit(node.lhs)
        self.plain_print(" == ", style="keyword")
        self.visit(node.rhs)
        self.plain_print(") ")

    def visit_Call(self, node: Call) -> None:
        self.plain_print(node.name)
        self.print_seq(node.args, delim=", ", prefix="(", suffix=")", emit=self.visit)

    def visit_BinOp(self, node: BinOp) -> None:
        self.plain_print("(")
        self.visit(node.lhs)
        self.plain_print(f" {node.op} ", style="keyword")
        self.visit(node.rhs)
        self.plain_print(")")

    def visit_UnaryOp(self, node: UnaryOp) -> None:
        self.plain_print(f"{node.op}", style="keyword")
        self.visit(node.operand)

    def visit_Bit(self, node: Bit) -> None:
        self.visit_Name(node.name)
        if node.addr is not None:
            self.plain_print("[")
            self.plain_print(node.addr, style="number")
            self.plain_print("]")

    def visit_Number(self, node: Number) -> None:
        self.plain_print(node.value)

    def visit_Pi(self, node: Pi) -> None:
        self.plain_print("pi", style="number")

    def visit_Name(self, node: Name) -> None:
        return self.plain_print(node.id, style="symbol")

    def visit_ParallelQArgs(self, node: ParallelQArgs) -> None:
        self.plain_print("{")
        with self.indent():
            for idx, qargs in enumerate(node.qargs):
                self.print_newline()
                self.print_seq(qargs, emit=self.visit)
                self.plain_print(";")
        self.print_newline()
        self.plain_print("}")

    def visit_ParaU3Gate(self, node: ParaU3Gate) -> None:
        self.plain_print("parallel.U", style="keyword")
        self.plain_print("(")
        self.visit(node.theta)
        self.plain_print(", ")
        self.visit(node.phi)
        self.plain_print(", ")
        self.visit(node.lam)
        self.plain_print(") ")
        self.visit_ParallelQArgs(node.qargs)

    def visit_ParaCZGate(self, node: ParaCZGate) -> None:
        self.plain_print("parallel.CZ ", style="keyword")
        self.visit_ParallelQArgs(node.qargs)

    def visit_ParaRZGate(self, node: ParaRZGate) -> None:
        self.plain_print("parallel.RZ", style="keyword")
        self.plain_print("(")
        self.visit(node.theta)
        self.plain_print(") ")
        self.visit_ParallelQArgs(node.qargs)

    def visit_GlobUGate(self, node: GlobUGate) -> None:
        self.plain_print("glob.U", style="keyword")
        self.plain_print("(")
        self.visit(node.theta)
        self.plain_print(", ")
        self.visit(node.phi)
        self.plain_print(", ")
        self.visit(node.lam)
        self.plain_print(") ")
        self.print_seq(node.registers, prefix="{", suffix="}", emit=self.visit)

    def visit_NoisePAULI1(self, node: NoisePAULI1) -> None:
        self.plain_print("noise.PAULI1", style="keyword")
        self.plain_print("(")
        self.visit(node.px)
        self.plain_print(", ")
        self.visit(node.py)
        self.plain_print(", ")
        self.visit(node.pz)
        self.plain_print(") ")
        self.visit(node.qarg)
        self.plain_print(";")



# %% [markdown]
# # GHZ State Preparation with Parallelism
# In this example, we will implement a *Greenberger-Horne-Zeilinger* (GHZ) state preparation circuit with $N = 2^n$ qubits.
#
# First, we will present the standard linear-depth construction in Bloqade but later we will present a log-depth
# construction that achieves the same result. We then take this one step further and use the fact that Bloqade
# (and QuEra's neutral atom hardware!) support *parallel* gates, allowing for the application of the same gate
# across multiple qubits simultaneously. Combined with the fact that atom *shuttling* allows for arbitrary
# connectivity, we can also decrease the circuit *execution* depth from $N$ to just $n$.

# %%
import math

from bloqade import qasm2
from kirin.dialects import ilist

# %% [markdown]
# ## Simple Linear Depth Implementation of a GHZ State Preparation Circuit
#
# A simple GHZ state preparation circuit can be built with $N - 1$ CX gates and $1$ H gate.
# This gives the circuit an execution depth of $N$.


# %%
def ghz_linear(n: int):
    n_qubits = int(2**n)

    @qasm2.extended
    def ghz_linear_program():

        qreg = qasm2.qreg(n_qubits)
        # Apply a Hadamard on the first qubit
        qasm2.h(qreg[0])
        # Create a cascading sequence of CX gates
        # necessary for quantum computers that
        # only have nearest-neighbor connectivity between qubits
        for i in range(1, n_qubits):
            qasm2.cx(qreg[i - 1], qreg[i])

    return ghz_linear_program


# %% [markdown]
# ## Log-depth Implementation of a GHZ State Preparation Circuit
#
# Let's take a look how we can rewrite the circuit to take advantage of QuEra's hardware capabilities.
# We can achieve log(N) circuit depth by [rearranging the CX gates (see *Mooney, White, Hill, Hollenberg* - 2021)](https://arxiv.org/abs/2101.08946).
#

# %% [markdown]
# <div class="admonition note">
# <p class="admonition-title">Circuit vs. Execution Depth</p>
# <p>
# Before going any further, it's worth distinguishing between the concept of circuit depth and circuit execution depth.
# For example, in the following implementation, each CX gate instruction inside the for-loop is executed in sequence.
# So even thought the circuit depth is $log(N) = n$, the circuit execution depth is still $N$.
# </p>
# </div>


# %%
def ghz_log_depth(n: int):
    n_qubits = int(2**n)

    @qasm2.extended
    def layer_of_cx(i_layer: int, qreg: qasm2.QReg):
        step = n_qubits // (2**i_layer)
        for j in range(0, n_qubits, step):
            qasm2.cx(ctrl=qreg[j], qarg=qreg[j + step // 2])

    @qasm2.extended
    def ghz_log_depth_program():

        qreg = qasm2.qreg(n_qubits)

        qasm2.h(qreg[0])
        for i in range(n):
            layer_of_cx(i_layer=i, qreg=qreg)

    return ghz_log_depth_program


# %% [markdown]
# ## Our Native Gate Set and Parallelism
# By nature, our digital quantum computer can execute native gates in parallel in an single instruction/ execution cycle.
# The concept is very similar to the SIMD (Single Instruction, Multiple Data) in classical computing.
#
# On our hardware, there are two important factors to be considered:
# 1. the native gate set allows for arbitrary (parallel) rotations and (parallel) CZ gates.
# 2. Our atom shuttling architecture allows arbitrary qubit connectivity. This means that our parallel instruction is not limited to fixed connectivity (for example nearest neighbor connectivity).
#
# With this in mind, we can rewrite the `layer` subroutine to now use the `qasm2.parallel` dialect in Bloqade.
# We know that the CX gate can be decomposed into a CZ gate with two single-qubit gates $R_y(-\pi/2)$ and $R_y(\pi/2)$ acting on the target qubits.
# With this decomposition in mind, we can now using our parallel gate instructions `parallel.u` and `parallel.cz`.
# With the following modification, we can further reduce the circuit execution depth to just $n$ (log of the total number of qubits $N$)


# %%
def ghz_log_simd(n: int):
    n_qubits = int(2**n)

    @qasm2.extended
    def layer(i_layer: int, qreg: qasm2.QReg):
        step = n_qubits // (2**i_layer)

        def get_qubit(x: int):
            return qreg[x]

        ctrl_qubits = ilist.Map(fn=get_qubit, collection=range(0, n_qubits, step))
        targ_qubits = ilist.Map(
            fn=get_qubit, collection=range(step // 2, n_qubits, step)
        )

        # Ry(-pi/2)
        qasm2.parallel.u(qargs=targ_qubits, theta=-math.pi / 2, phi=0.0, lam=0.0)

        # CZ gates
        qasm2.parallel.cz(ctrls=ctrl_qubits, qargs=targ_qubits)

        # Ry(pi/2)
        qasm2.parallel.u(qargs=targ_qubits, theta=math.pi / 2, phi=0.0, lam=0.0)

    @qasm2.extended
    def ghz_log_depth_program():

        qreg = qasm2.qreg(n_qubits)

        qasm2.h(qreg[0])
        for i in range(n):
            layer(i_layer=i, qreg=qreg)

    return ghz_log_depth_program


# %% [markdown]
# <div class="admonition warning">
# <p class="admonition-title">Using Closures to Capture Global Variables</p>
# <p>
# While bloqade.qasm2 permits a main program with arguments, standard QASM2 does not.
# To get around this, we need to put the program in a closure.
# Our Kirin compiler toolchain can capture the global variable inside the closure.
# In this case, the n_qubits will be captured upon calling the ghz_log_simd(n_qubits) python function.
# As a result, the returned QASM2 program will not have any arguments.
# </p>
# </div>



from .groups import main as main
from ._wrappers import *  # noqa: F403
from .dialects.aux import *  # noqa F403
from .dialects.gate import *  # noqa F403
from .dialects.noise import *  # noqa F403
from .dialects.collapse import *  # noqa F403



import json

import numpy as np
import pytest
from bloqade.visual.animation.runtime import utils


def test_json_numpy():
    original_array = np.random.uniform(size=(1, 5, 3, 2, 4))
    array_json = utils.array_to_json(original_array)
    # serialize then deserialize to ensure the json is in the correct format
    obj = json.loads(json.dumps(array_json))
    # unpack the json object
    unpacked_array = utils.json_to_array(obj)
    # check that the array is the same bytes
    assert unpacked_array.tobytes() == original_array.tobytes()


def test_validation():
    a = np.array([((1, 2), 2), ((3, 4), 4)], dtype=[("x", "O", (2,)), ("y", "i8")])

    with pytest.raises(ValueError):
        utils.array_to_json(a)



from kirin.emit import EmitStrFrame
from kirin.interp import MethodTable, impl
from bloqade.stim.emit.stim import EmitStimMain

from . import stmts
from ._dialect import dialect
from .stmts.reset import Reset
from .stmts.measure import Measurement


@dialect.register(key="emit.stim")
class EmitStimCollapseMethods(MethodTable):

    meas_map: dict[str, str] = {
        stmts.MX.name: "MX",
        stmts.MY.name: "MY",
        stmts.MZ.name: "MZ",
        stmts.MXX.name: "MXX",
        stmts.MYY.name: "MYY",
        stmts.MZZ.name: "MZZ",
    }

    @impl(stmts.MX)
    @impl(stmts.MY)
    @impl(stmts.MZ)
    @impl(stmts.MXX)
    @impl(stmts.MYY)
    @impl(stmts.MZZ)
    def get_measure(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: Measurement):

        probability: str = frame.get(stmt.p)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        out = f"{self.meas_map[stmt.name]}({probability}) " + " ".join(targets)
        emit.writeln(frame, out)

        return ()

    reset_map: dict[str, str] = {
        stmts.RX.name: "RX",
        stmts.RY.name: "RY",
        stmts.RZ.name: "RZ",
    }

    @impl(stmts.RX)
    @impl(stmts.RY)
    @impl(stmts.RZ)
    def get_reset(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: Reset):

        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        out = f"{self.reset_map[stmt.name]} " + " ".join(targets)
        emit.writeln(frame, out)

        return ()

    @impl(stmts.PPMeasurement)
    def pp_measure(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.PPMeasurement
    ):
        probability: str = frame.get(stmt.p)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        out = f"MPP({probability}) " + " ".join(targets)
        emit.writeln(frame, out)

        return ()



import abc
from typing import Dict, List, Tuple, Iterable
from dataclasses import dataclass

from kirin import ir
from kirin.rewrite import abc as rewrite_abc, result
from kirin.dialects import ilist
from bloqade.analysis import address
from kirin.analysis.const import lattice
from bloqade.qasm2.dialects import uop, core, parallel
from bloqade.analysis.schedule import StmtDag


class MergePolicyABC(abc.ABC):
    @abc.abstractmethod
    def __call__(self, node: ir.Statement) -> result.RewriteResult:
        pass

    @classmethod
    @abc.abstractmethod
    def can_merge(cls, stmt1: ir.Statement, stmt2: ir.Statement) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def merge_gates(
        cls, gate_stmts: Iterable[ir.Statement]
    ) -> List[List[ir.Statement]]:
        pass

    @classmethod
    @abc.abstractmethod
    def from_analysis(
        cls, dag: StmtDag, address_analysis: Dict[ir.SSAValue, address.Address]
    ) -> "MergePolicyABC":
        pass


@dataclass
class SimpleMergePolicy(MergePolicyABC):
    """General merge policy for merging gates based on their type and arguments.

    Base class to implement a merge policy for CZ, U and RZ gates, To completed the policy implement the
    `merge_gates` class method. This will take an iterable of statements and return a list
    of groups of statements that can be merged together. There are two mix-in classes
    that can be used to implement the `merge_gates` method. The `GreedyMixin` will merge
    gates together greedily, while the `OptimalMixIn` will merge gates together optimally.

    """

    address_analysis: Dict[ir.SSAValue, address.Address]
    """Mapping from SSA values to their address analysis results. Needed for rewrites"""
    merge_groups: List[List[ir.Statement]]
    """List of groups of statements that can be merged together"""
    group_numbers: Dict[ir.Statement, int]
    """Mapping from statements to their group number"""

    @staticmethod
    def same_id_checker(ssa1: ir.SSAValue, ssa2: ir.SSAValue):
        if ssa1 is ssa2:
            return True
        elif (hint1 := ssa1.hints.get("const")) and (hint2 := ssa2.hints.get("const")):
            assert isinstance(hint1, lattice.Result) and isinstance(
                hint2, lattice.Result
            )
            return hint1.is_equal(hint2)
        else:
            return False

    @classmethod
    def check_equiv_args(
        cls,
        args1: Iterable[ir.SSAValue],
        args2: Iterable[ir.SSAValue],
    ):
        try:
            return all(
                cls.same_id_checker(ssa1, ssa2)
                for ssa1, ssa2 in zip(args1, args2, strict=True)
            )
        except ValueError:
            return False

    @classmethod
    def can_merge(cls, stmt1: ir.Statement, stmt2: ir.Statement) -> bool:
        match stmt1, stmt2:
            case (
                (uop.UGate(), uop.UGate())
                | (uop.RZ(), uop.RZ())
                | (parallel.UGate(), parallel.UGate())
                | (parallel.UGate(), uop.UGate())
                | (uop.UGate(), parallel.UGate())
                | (uop.UGate(), parallel.UGate())
                | (uop.UGate(), parallel.UGate())
                | (parallel.RZ(), parallel.RZ())
                | (uop.RZ(), parallel.RZ())
                | (parallel.RZ(), uop.RZ())
            ):

                return cls.check_equiv_args(stmt1.args[1:], stmt2.args[1:])
            case (
                (parallel.CZ(), parallel.CZ())
                | (parallel.CZ(), uop.CZ())
                | (uop.CZ(), parallel.CZ())
                | (uop.CZ(), uop.CZ())
            ):
                return True

            case _:
                return False

    @classmethod
    def from_analysis(
        cls,
        dag: StmtDag,
        address_analysis: Dict[ir.SSAValue, address.Address],
    ):

        merge_groups = []
        group_numbers = {}

        for group in dag.topological_groups():

            gate_groups = cls.merge_gates(map(dag.stmts.__getitem__, group))

            gate_groups_iter = (group for group in gate_groups if len(group) > 1)

            for gate_group in gate_groups_iter:
                group_number = len(merge_groups)
                merge_groups.append(gate_group)
                for stmt in gate_group:
                    group_numbers[stmt] = group_number

        return cls(
            address_analysis=address_analysis,
            merge_groups=merge_groups,
            group_numbers=group_numbers,
        )

    def __call__(self, node: ir.Statement) -> result.RewriteResult:
        if node not in self.group_numbers:
            return result.RewriteResult()

        group = self.merge_groups[self.group_numbers[node]]
        if node is group[0]:
            method = getattr(self, f"rewrite_group_{node.name}")
            method(node, group)

        node.delete()

        return result.RewriteResult(has_done_something=True)

    def move_and_collect_qubit_list(
        self, node: ir.Statement, qargs: List[ir.SSAValue]
    ) -> Tuple[ir.SSAValue, ...]:

        qubits = []
        # collect references to qubits
        for qarg in qargs:
            addr = self.address_analysis[qarg]

            if isinstance(addr, address.AddressQubit):
                qubits.append(qarg)

            elif isinstance(addr, address.AddressTuple):
                assert isinstance(qarg, ir.ResultValue)
                assert isinstance(qarg.stmt, ilist.New)
                qubits.extend(qarg.stmt.values)

        # for qubits coming from QRegGet, move both the get statement
        # and the index statement before the current node, we do not need
        # to move the register because the current node has some dependency
        # on it.
        for qarg in qubits:
            if (
                isinstance(qarg, ir.ResultValue)
                and isinstance(qarg.owner, core.QRegGet)
                and isinstance(qarg.owner.idx, ir.ResultValue)
            ):
                idx = qarg.owner.idx
                idx.owner.delete(safe=False)
                idx.owner.insert_before(node)
                qarg.owner.delete(safe=False)
                qarg.owner.insert_before(node)

        return tuple(qubits)

    def rewrite_group_cz(self, node: ir.Statement, group: List[ir.Statement]):
        ctrls = []
        qargs = []

        for stmt in group:
            if isinstance(stmt, uop.CZ):
                ctrls.append(stmt.ctrl)
                qargs.append(stmt.qarg)
            elif isinstance(stmt, parallel.CZ):
                ctrls.append(stmt.ctrls)
                qargs.append(stmt.qargs)
            else:
                raise RuntimeError(f"Unexpected statement {stmt}")

        ctrls_values = self.move_and_collect_qubit_list(node, ctrls)
        qargs_values = self.move_and_collect_qubit_list(node, qargs)

        new_ctrls = ilist.New(values=ctrls_values)
        new_qargs = ilist.New(values=qargs_values)
        new_gate = parallel.CZ(ctrls=new_ctrls.result, qargs=new_qargs.result)

        new_ctrls.insert_before(node)
        new_qargs.insert_before(node)
        new_gate.insert_before(node)

    def rewrite_group_U(self, node: ir.Statement, group: List[ir.Statement]):
        self.rewrite_group_u(node, group)

    def rewrite_group_u(self, node: ir.Statement, group: List[ir.Statement]):
        qargs = []

        for stmt in group:
            if isinstance(stmt, uop.UGate):
                qargs.append(stmt.qarg)
            elif isinstance(stmt, parallel.UGate):
                qargs.append(stmt.qargs)
            else:
                raise RuntimeError(f"Unexpected statement {stmt}")

        assert isinstance(node, (uop.UGate, parallel.UGate))

        qargs_values = self.move_and_collect_qubit_list(node, qargs)

        new_qargs = ilist.New(values=qargs_values)
        new_gate = parallel.UGate(
            qargs=new_qargs.result,
            theta=node.theta,
            phi=node.phi,
            lam=node.lam,
        )
        new_qargs.insert_before(node)
        new_gate.insert_before(node)

    def rewrite_group_rz(self, node: ir.Statement, group: List[ir.Statement]):
        qargs = []

        for stmt in group:
            if isinstance(stmt, uop.RZ):
                qargs.append(stmt.qarg)
            elif isinstance(stmt, parallel.RZ):
                qargs.append(stmt.qargs)
            else:
                raise RuntimeError(f"Unexpected statement {stmt}")

        assert isinstance(node, (uop.RZ, parallel.RZ))

        qargs_values = self.move_and_collect_qubit_list(node, qargs)
        new_qargs = ilist.New(values=qargs_values)
        new_gate = parallel.RZ(
            qargs=new_qargs.result,
            theta=node.theta,
        )
        new_qargs.insert_before(node)
        new_gate.insert_before(node)


class GreedyMixin(MergePolicyABC):
    """Merge policy that greedily merges gates together.

    The `merge_gates` method will merge policy will try greedily merge gates together.
    This policy has a worst case complexity of O(n) where n is the
    number of gates in the input iterable.
    """

    @classmethod
    def merge_gates(
        cls, gate_stmts: Iterable[ir.Statement]
    ) -> List[List[ir.Statement]]:

        iterable = iter(gate_stmts)
        groups = [[next(iterable)]]

        for stmt in gate_stmts:
            if cls.can_merge(groups[-1][-1], stmt):
                groups[-1].append(stmt)
            else:
                groups.append([stmt])

        return groups


class OptimalMixIn(MergePolicyABC):
    """Merge policy that merges gates together optimally.

    The `merge_gates` method will merge policy will try to merge every gate into every
    group of gates, terminating when it finds a group that can be merged with the current
    gate. This policy has a worst case complexity of O(n^2) where n is the number of gates
    in the input iterable.

    """

    @classmethod
    def merge_gates(
        cls, gate_stmts: Iterable[ir.Statement]
    ) -> List[List[ir.Statement]]:

        groups = []
        for stmt in gate_stmts:
            found = False
            for group in groups:
                if cls.can_merge(group[-1], stmt):
                    group.append(stmt)
                    found = True
                    break

            if not found:
                groups.append([stmt])

        return groups


@dataclass
class SimpleGreedyMergePolicy(GreedyMixin, SimpleMergePolicy):
    pass


@dataclass
class SimpleOptimalMergePolicy(OptimalMixIn, SimpleMergePolicy):
    pass


@dataclass
class UOpToParallelRule(rewrite_abc.RewriteRule):
    merge_rewriters: Dict[ir.Block | None, MergePolicyABC]

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        merge_rewriter = self.merge_rewriters.get(
            node.parent_block, lambda _: result.RewriteResult()
        )
        return merge_rewriter(node)



from typing import List

from kirin import ir, types
from bloqade import qasm2
from kirin.dialects import py, func
from bloqade.qasm2.passes.parallel import ParallelToUOp


def as_int(value: int):
    return py.constant.Constant(value=value)


def as_float(value: float):
    return py.constant.Constant(value=value)


def test_cz_rewrite():

    @qasm2.extended
    def main():
        q = qasm2.qreg(4)

        qasm2.parallel.cz(ctrls=[q[0], q[2]], qargs=[q[1], q[3]])

    # Run rewrite
    ParallelToUOp(main.dialects)(main)

    # post-rewrite expected function
    expected: List[ir.Statement] = [
        (n_qubits := as_int(4)),
        (reg := qasm2.core.QRegNew(n_qubits=n_qubits.result)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg.result, idx=idx0.result)),
        (idx2 := as_int(2)),
        (q2 := qasm2.core.QRegGet(reg.result, idx=idx2.result)),
        (idx1 := as_int(1)),
        (q1 := qasm2.core.QRegGet(reg.result, idx=idx1.result)),
        (idx3 := as_int(3)),
        (q3 := qasm2.core.QRegGet(reg.result, idx=idx3.result)),
        (qasm2.uop.CZ(ctrl=q0.result, qarg=q1.result)),
        (qasm2.uop.CZ(ctrl=q2.result, qarg=q3.result)),
        (return_none := func.ConstantNone()),
        (func.Return(return_none.result)),
    ]
    block = ir.Block(expected)
    block.args.append_from(types.MethodType[[], types.NoneType], "main_self")
    expected_func_stmt = func.Function(
        sym_name="main",
        signature=func.Signature(inputs=(), output=types.NoneType),
        body=ir.Region(blocks=block),
    )

    expected_method = ir.Method(
        mod=None,
        py_func=None,
        sym_name="main",
        dialects=qasm2.main,
        code=expected_func_stmt,
        arg_names=[],
    )

    qasm2.main.run_pass(expected_method)  # type: ignore

    try:
        assert expected_method.code.is_equal(main.code)
    except AssertionError as e:
        print("Expected:")
        expected_method.print()
        print("Actual:")
        main.print()
        raise e


if __name__ == "__main__":
    test_cz_rewrite()



from typing import Dict, List, Optional
from dataclasses import dataclass

from kirin import ir
from kirin.rewrite import abc, result
from bloqade.analysis import address
from bloqade.qasm2.dialects import uop, parallel


@dataclass
class ParallelToUOpRule(abc.RewriteRule):
    id_map: Dict[int, ir.SSAValue]
    address_analysis: Dict[ir.SSAValue, address.Address]

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if node.dialect == parallel.dialect:
            return getattr(self, f"rewrite_{node.name}")(node)

        return result.RewriteResult()

    def get_qubit_ssa(self, ilist_ref: ir.SSAValue) -> Optional[List[ir.SSAValue]]:
        addr = self.address_analysis.get(ilist_ref)
        if not isinstance(addr, address.AddressTuple):
            return None

        ids = []
        for ele in addr.data:
            if not isinstance(ele, address.AddressQubit):
                return None

            ids.append(ele.data)

        return [self.id_map[ele] for ele in ids]

    def rewrite_cz(self, node: ir.Statement):
        assert isinstance(node, parallel.CZ)

        ctrls = self.get_qubit_ssa(node.ctrls)
        qargs = self.get_qubit_ssa(node.qargs)

        if ctrls is None or qargs is None:
            return result.RewriteResult()

        for ctrl, qarg in zip(ctrls, qargs):
            new_node = uop.CZ(ctrl, qarg)
            new_node.insert_before(node)

        node.delete()

        return result.RewriteResult(has_done_something=True)

    def rewrite_u(self, node: ir.Statement):
        assert isinstance(node, parallel.UGate)

        qargs = self.get_qubit_ssa(node.qargs)

        if qargs is None:
            return result.RewriteResult()

        for qarg in qargs:
            new_node = uop.UGate(qarg, theta=node.theta, phi=node.phi, lam=node.lam)
            new_node.insert_after(node)

        node.delete()

        return result.RewriteResult(has_done_something=True)

    def rewrite_rz(self, node: ir.Statement):
        assert isinstance(node, parallel.RZ)

        qargs = self.get_qubit_ssa(node.qargs)

        if qargs is None:
            return result.RewriteResult()

        for qarg in qargs:
            new_node = uop.RZ(qarg, theta=node.theta)
            new_node.insert_after(node)

        node.delete()

        return result.RewriteResult(has_done_something=True)



import itertools
import dataclasses
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, List, Tuple, Callable, Optional

import matplotlib.patches as mpatches
import matplotlib.collections as mcollections

from .gate_event import GateEvent

_gate_color_code = {
    "GlobalCZGate": "tab:red",
    "TopHatCZGate": "tab:red",
    "GlobalHyperfineRotation": "tab:blue",
    "GlobalHyperfineZRotation": "tab:green",
    "LocalHyperfineZRotation": "tab:green",
    "LocalHyperfineZRotationPos": "tab:green",
    "LocalHyperfineRotation": "tab:blue",
    "LocalHyperfineRotationPos": "tab:blue",
}


class quera_color_code(str, Enum):
    # TODO: for python 3.11+, replace traits with StrEnum
    purple = "#6437FF"
    red = "#C2477F"
    yellow = "#EADD08"


@dataclasses.dataclass
class GateArtist:
    mpl_ax: Any

    @abstractmethod
    def clear_data(self) -> None:
        pass

    @abstractmethod
    def get_artists(self) -> Tuple[Any]:
        pass


@dataclasses.dataclass(init=False)
class GlobalGateArtist(GateArtist):
    mpl_obj: mpatches.Rectangle

    def __init__(self, mpl_ax: Any, xmin, ymin, width, height, color):
        super().__init__(mpl_ax)
        rc = mpatches.Rectangle(
            [xmin, ymin], width, height, color=color, alpha=0.6, visible=False
        )
        mpl_ax.add_patch(rc)
        self.mpl_obj = rc

    def clear_data(self) -> None:
        self.mpl_obj.set_width(0)
        self.mpl_obj.set_height(0)
        self.mpl_obj.set_xy([0, 0])

    def get_artists(self) -> Tuple[Any]:
        return (self.mpl_obj,)

    def set_visible(self, visible: bool):
        self.mpl_obj.set_visible(visible)


@dataclasses.dataclass(init=False)
class RowRegionGateArtist(GateArtist):
    """
    A row region gate artist object.

    bound box is [y_origin - width/2, y_origin + width/2]
    """

    mpl_obj: mpatches.Rectangle
    mpl_obj_keepout_top: mpatches.Rectangle
    mpl_obj_keepout_btm: mpatches.Rectangle
    width: float
    xmin: float

    def __init__(
        self, mpl_ax: Any, xmin, width, ymin, ymin_keepout, ymax, ymax_keepout, color
    ):
        super().__init__(mpl_ax)
        self.width = width
        self.xmin = xmin
        rc_btm = mpatches.Rectangle(
            [xmin, ymin_keepout],
            width,
            ymin - ymin_keepout,
            color=color,
            alpha=0.3,
            visible=False,
        )
        mpl_ax.add_patch(rc_btm)
        self.mpl_obj_keepout_btm = rc_btm

        rc = mpatches.Rectangle(
            [xmin, ymin], width, ymax - ymin, color=color, alpha=0.6, visible=False
        )
        mpl_ax.add_patch(rc)
        self.mpl_obj = rc

        rc_top = mpatches.Rectangle(
            [xmin, ymax],
            width,
            ymax_keepout - ymax,
            color=color,
            alpha=0.3,
            visible=False,
        )
        mpl_ax.add_patch(rc_top)
        self.mpl_obj_keepout_top = rc_top

    def clear_data(self) -> None:
        self.mpl_obj.set_width(0)
        self.mpl_obj.set_height(0)
        self.mpl_obj.set_xy([0, 0])

        self.mpl_obj_keepout_top.set_width(0)
        self.mpl_obj_keepout_top.set_height(0)
        self.mpl_obj_keepout_top.set_xy([0, 0])

        self.mpl_obj_keepout_btm.set_width(0)
        self.mpl_obj_keepout_btm.set_height(0)
        self.mpl_obj_keepout_btm.set_xy([0, 0])

    def get_artists(self) -> Tuple[Any]:
        return (self.mpl_obj, self.mpl_obj_keepout_top, self.mpl_obj_keepout_btm)

    def update_data(self, ymin, ymax, ymin_keepout, ymax_keepout):
        self.mpl_obj.set_height(ymax - ymin)
        self.mpl_obj.set_width(self.width)
        self.mpl_obj.set_xy([self.xmin, ymin])

        self.mpl_obj_keepout_top.set_height(ymax_keepout - ymax)
        self.mpl_obj_keepout_top.set_width(self.width)
        self.mpl_obj_keepout_top.set_xy([self.xmin, ymax])

        self.mpl_obj_keepout_btm.set_height(ymin - ymin_keepout)
        self.mpl_obj_keepout_btm.set_width(self.width)
        self.mpl_obj_keepout_btm.set_xy([self.xmin, ymin_keepout])

    def set_visible(self, visible: bool):
        self.mpl_obj.set_visible(visible)
        self.mpl_obj_keepout_btm.set_visible(visible)
        self.mpl_obj_keepout_top.set_visible(visible)


@dataclasses.dataclass
class SpotGateArtist(GateArtist):
    mpl_obj: mcollections.PathCollection

    def __init__(self, mpl_ax: Any, xy: List[Tuple[float, float]], radius, color):
        super().__init__(mpl_ax)
        self.mpl_obj = mpl_ax.scatter(
            [x for x, y in xy],
            [y for x, y in xy],
            s=radius,
            visible=False,
            marker="o",
            facecolors=color,
            alpha=0.5,
            zorder=+100,
        )

    def clear_data(self) -> None:
        self.mpl_obj.set_offsets([])

    def get_artists(self) -> Tuple[Any]:
        return (self.mpl_obj,)

    def update_data(self, xy: List[Tuple[float, float]]):
        self.mpl_obj.set_offsets(xy)

    def set_visible(self, visible: bool):
        self.mpl_obj.set_visible(visible)


@dataclasses.dataclass
class FieldOfView:
    xmin: Optional[float] = dataclasses.field(default=None)
    xmax: Optional[float] = dataclasses.field(default=None)
    ymin: Optional[float] = dataclasses.field(default=None)
    ymax: Optional[float] = dataclasses.field(default=None)

    def not_defined(self):
        return (
            self.xmin is None
            or self.xmax is None
            or self.ymin is None
            or self.ymax is None
        )

    @property
    def width(self):
        return self.xmax - self.xmin

    @property
    def height(self):
        return self.ymax - self.ymin

    def to_json(self):
        return {
            "xmin": self.xmin,
            "xmax": self.xmax,
            "ymin": self.ymin,
            "ymax": self.ymax,
        }

    @classmethod
    def from_json(cls, json_dict):
        return FieldOfView(
            xmin=json_dict["xmin"],
            xmax=json_dict["xmax"],
            ymin=json_dict["ymin"],
            ymax=json_dict["ymax"],
        )


@dataclasses.dataclass(init=False)
class GatePainter:
    artist_objs: Dict[str, GateArtist]
    artist_methods: Dict[str, Callable]

    def __init__(self, mpl_ax, qpu_fov: FieldOfView, scale: float = 1.0):
        # gates:
        self.artist_objs = {}
        self.artist_methods = {
            "GlobalCZGate": self.process_global_cz_gate,
            "GlobalHyperfineRotation": self.process_global_hyperfine_rotation,
            "TopHatCZGate": self.process_tophat_cz_gate,
            "LocalHyperfineZRotation": self.process_local_hyperfine_z_rotation,
            "LocalHyperfineZRotationPos": self.process_local_hyperfine_z_rotation_pos,
            "LocalHyperfineRotation": self.process_local_hyperfine_rotation,
            "LocalHyperfineRotationPos": self.process_local_hyperfine_rotation_pos,
        }
        self.artist_objs["GlobalCZGate"] = GlobalGateArtist(
            mpl_ax=mpl_ax,
            xmin=qpu_fov.xmin,
            width=qpu_fov.width,
            ymin=qpu_fov.ymin,
            height=qpu_fov.height,
            color=_gate_color_code["GlobalCZGate"],
        )
        self.artist_objs["GlobalHyperfineRotation"] = GlobalGateArtist(
            mpl_ax=mpl_ax,
            xmin=qpu_fov.xmin,
            width=qpu_fov.width,
            ymin=qpu_fov.ymin,
            height=qpu_fov.height,
            color=_gate_color_code["GlobalHyperfineRotation"],
        )
        self.artist_objs["TopHatCZGate"] = RowRegionGateArtist(
            mpl_ax=mpl_ax,
            xmin=qpu_fov.xmin,
            width=qpu_fov.width,
            ymin=0,
            ymin_keepout=0,
            ymax=0,
            ymax_keepout=0,
            color=_gate_color_code["TopHatCZGate"],
        )

        self.artist_objs["LocalHyperfineZRotation"] = SpotGateArtist(
            mpl_ax=mpl_ax,
            xy=[],
            radius=160 * scale,
            color=_gate_color_code["LocalHyperfineZRotation"],
        )
        self.artist_objs["LocalHyperfineZRotationPos"] = SpotGateArtist(
            mpl_ax=mpl_ax,
            xy=[],
            radius=160 * scale,
            color=_gate_color_code["LocalHyperfineZRotationPos"],
        )

        self.artist_objs["LocalHyperfineRotation"] = SpotGateArtist(
            mpl_ax=mpl_ax,
            xy=[],
            radius=80,
            color=_gate_color_code["LocalHyperfineRotation"],
        )
        self.artist_objs["LocalHyperfineRotationPos"] = SpotGateArtist(
            mpl_ax=mpl_ax,
            xy=[],
            radius=80,
            color=_gate_color_code["LocalHyperfineRotationPos"],
        )

    @staticmethod
    def process_global_cz_gate(artist_obj, **kwargs) -> None:
        artist_obj.set_visible(True)

    @staticmethod
    def process_global_hyperfine_rotation(artist_obj, **kwargs) -> None:
        artist_obj.set_visible(True)

    @staticmethod
    def process_tophat_cz_gate(artist_obj, **kwargs) -> None:
        artist_obj.update_data(
            ymin=kwargs["y_min"],
            ymax=kwargs["y_max"],
            ymin_keepout=kwargs["y_min_keepout"],
            ymax_keepout=kwargs["y_max_keepout"],
        )
        artist_obj.set_visible(True)

    @staticmethod
    def process_local_hyperfine_z_rotation_pos(artist_obj, **kwargs) -> None:
        artist_obj.update_data(xy=[(x, kwargs["y_row_loc"]) for x in kwargs["varargs"]])
        artist_obj.set_visible(True)

    @staticmethod
    def process_local_hyperfine_rotation_pos(artist_obj, **kwargs) -> None:
        artist_obj.update_data(xy=[(x, kwargs["y_row_loc"]) for x in kwargs["varargs"]])
        artist_obj.set_visible(True)

    @staticmethod
    def process_local_hyperfine_z_rotation(artist_obj, **kwargs) -> None:
        raise ValueError(
            "cannot have local HyperfineZRotation, it should be already lowered to Pos variant!"
        )

    @staticmethod
    def process_local_hyperfine_rotation(artist_obj, **kwargs) -> None:
        raise ValueError(
            "cannot have local HyperfineRotation, it should be already lowered to Pos variant!"
        )

    def process_gates(self, gates: List[GateEvent]) -> List:

        for gate_artist in self.artist_objs.values():
            gate_artist.set_visible(False)

        for gate in gates:
            self.artist_methods[gate.cls_name](
                self.artist_objs[gate.cls_name], **gate.kwargs
            )

        return list(
            itertools.chain.from_iterable(
                gate_artist.get_artists() for gate_artist in self.artist_objs.values()
            )
        )



# %% [markdown]
# # Repeat Until Success with STAR Gadget
# In this example, we will demonstrate a near-term fault tolerant gadget
# which is a repeat-until-success protocol to implement a Z phase gate
# using a resource state (similar to a T state), Pauli gates, and feed-forward measurement.
#
# For more information, see https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.010337,
# especially Fig. 7.

# %%
from bloqade import qasm2

# %% [markdown]
# This example highlights a few interesting capabilities of having a full kernel structure with
# runtime control flow. One example is the ability to dynamically allocate qubits, possibly
# based on previous run-time measurement outcomes.
#
# In this case, we prepare a resource state, which is a generalization of the T state with
# an arbitrary Z rotation $|0\rangle + e^{i\theta}|1\rangle$.


# %%
@qasm2.extended
def prep_resource_state(theta: float):
    qreg = qasm2.qreg(1)
    qubit = qreg[0]
    qasm2.h(qubit)
    qasm2.rz(qubit, theta)
    return qubit


# %% [markdown]
# Using this resource state, we can teleport the Z phase gate to a target qubit using
# only Clifford gates, which are much easier to implement fault-tolerantly.
# This is implemented by first applying a CNOT gate controlled by the resource state
# on the target qubit, then measuring the target qubit in the computational basis.
# If the measurement outcome is 1 (which occurs with 50% probability), the gadget
# executed a Z(theta) gate on the target qubit and teleported it
# to the new resource state.
#
# However, if the measurement outcome is 0 (which occurs with 50% probability),
# we apply an X gate, and the gadget executed a Z(-theta) gate on the target qubit.
# In order to correct this gate, we must apply a Z(+2*theta) gate on the new target state.
# Of course, we can apply this Z(+2*theta) gate by applying the same gadget with twice
# the angle, and repeat until we get the correct outcome.

# %% [markdown]
# The simplest way to implement the gadget is to simply post-select the correct measurement outcome
# using an assert statement. This is straightforward, but comes with an exponential overhead in the
# number of resource states: there is a 50% chance of success at each step, so there is only a
# $2^{-n}$ chance of success after $n$ Z phase gates.


# %%
@qasm2.extended
def z_phase_gate_postselect(target: qasm2.Qubit, theta: float):
    ancilla = prep_resource_state(theta)
    qasm2.cx(ancilla, target)
    creg = qasm2.creg(1)
    qasm2.measure(target, creg[0])


# %% [markdown]
# To (deterministically) implement the gate, we can recursively apply the gadget.
# Observe that, while it is efficient to represent this as a composition of kernels,
# there is no equivalent representation as a circuit, as the number of resource qubits and
# total number of gates is not known until runtime.


# %%
@qasm2.extended
def z_phase_gate_recursive(target: qasm2.Qubit, theta: float):
    """
    https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.010337 Fig. 7
    """
    ancilla = prep_resource_state(theta)
    qasm2.cx(ancilla, target)
    creg = qasm2.creg(1)
    qasm2.measure(target, creg[0])
    if creg[0] == 1:
        qasm2.reset(ancilla)
        z_phase_gate_recursive(ancilla, 2 * theta)
        return


# %% [markdown]
# An alternative representation uses control flow to
# implement the same gate. If the number of repeats is fixed, this can be represented
# as a static circuit, though it would require a large number of resource qubits and
# may still fail with a small probability $2^{-attempts}$.


# %%
@qasm2.extended
def z_phase_gate_loop(target: qasm2.Qubit, theta: float, attempts: int):
    """
    https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.010337 Fig. 7
    """
    creg = qasm2.creg(1)

    for ctr in range(attempts):
        ancilla = prep_resource_state(theta * (2**ctr))
        if creg[0] == 0:
            qasm2.cx(ancilla, target)
            qasm2.measure(target, creg[0])

        if creg[0] == 1:
            qasm2.reset(ancilla)
            target = ancilla


# Lets try to get pyqrack interpreter to run

# %%
from bloqade.pyqrack import PyQrack  # noqa: E402

theta = 0.1


@qasm2.extended
def recursion_main():
    target = qasm2.qreg(1)
    z_phase_gate_recursive(target[0], theta)
    return target


@qasm2.extended
def loop_main():
    target = qasm2.qreg(1)
    z_phase_gate_loop(target[0], theta, 5)
    return target


device = PyQrack()
qreg = device.run(recursion_main)
print(qreg)

# %% [markdown]
# Lets unwrap the postselection and loop versions of the gadget to see the qasm circuit.

# %%
from bloqade.qasm2.emit import QASM2  # noqa: E402
from bloqade.qasm2.parse import pprint  # noqa: E402

# qasm2 does not support parameterized circuits, so the entry point must be a function
# of zero arguments. We can use a closure to pass the parameters from outside the function.

# %%
theta = 0.1


@qasm2.extended
def main():
    target = qasm2.qreg(1)
    return z_phase_gate_postselect(target[0], theta)


target = QASM2()
ast = target.emit(main)
pprint(ast)

# %% [markdown]
# And now the loop version, which first needs to be unwrapped:

# %%
pprint(target.emit(loop_main))



from bloqade import stim

from .base import codegen


def test_obs_inc():

    @stim.main
    def test_simple_obs_inc():
        stim.ObservableInclude(idx=3, targets=(stim.GetRecord(-3), stim.GetRecord(-1)))

    out = codegen(test_simple_obs_inc)

    assert out.strip() == "OBSERVABLE_INCLUDE(3) rec[-3] rec[-1]"



from kirin.emit import EmitStrFrame
from kirin.interp import MethodTable, impl
from bloqade.stim.emit.stim import EmitStimMain

from . import stmts
from ._dialect import dialect
from .stmts.base import SingleQubitGate, ControlledTwoQubitGate


@dialect.register(key="emit.stim")
class EmitStimGateMethods(MethodTable):

    gate_1q_map: dict[str, tuple[str, str]] = {
        stmts.X.name: ("X", "X"),
        stmts.Y.name: ("Y", "Y"),
        stmts.Z.name: ("Z", "Z"),
        stmts.H.name: ("H", "H"),
        stmts.S.name: ("S", "S_DAG"),
        stmts.SqrtX.name: ("SQRT_X", "SQRT_X_DAG"),
        stmts.SqrtY.name: ("SQRT_Y", "SQRT_Y_DAG"),
        stmts.SqrtZ.name: ("SQRT_Z", "SQRT_Z_DAG"),
    }

    @impl(stmts.X)
    @impl(stmts.Y)
    @impl(stmts.Z)
    @impl(stmts.S)
    @impl(stmts.H)
    @impl(stmts.SqrtX)
    @impl(stmts.SqrtY)
    @impl(stmts.SqrtZ)
    def single_qubit_gate(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: SingleQubitGate
    ):
        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        res = f"{self.gate_1q_map[stmt.name][int(stmt.dagger)]} " + " ".join(targets)
        emit.writeln(frame, res)

        return ()

    gate_2q_map: dict[str, tuple[str, str]] = {
        stmts.Swap.name: ("SWAP", "SWAP"),
    }

    @impl(stmts.Swap)
    def two_qubit_gate(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: ControlledTwoQubitGate
    ):
        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        res = f"{self.gate_ctrl_2q_map[stmt.name][int(stmt.dagger)]} " + " ".join(
            targets
        )
        emit.writeln(frame, res)

        return ()

    gate_ctrl_2q_map: dict[str, tuple[str, str]] = {
        stmts.CX.name: ("CX", "CX"),
        stmts.CY.name: ("CY", "CY"),
        stmts.CZ.name: ("CZ", "CZ"),
        stmts.Swap.name: ("SWAP", "SWAP"),
    }

    @impl(stmts.CX)
    @impl(stmts.CY)
    @impl(stmts.CZ)
    def ctrl_two_qubit_gate(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: ControlledTwoQubitGate
    ):
        controls: tuple[str, ...] = frame.get_values(stmt.controls)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        res = f"{self.gate_ctrl_2q_map[stmt.name][int(stmt.dagger)]} " + " ".join(
            f"{ctrl} {tgt}" for ctrl, tgt in zip(controls, targets)
        )
        emit.writeln(frame, res)

        return ()

    @impl(stmts.SPP)
    def spp(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.SPP):

        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        res = "SPP " + " ".join(targets)
        emit.writeln(frame, res)

        return ()



from bloqade.types import Qubit as Qubit, QubitType as QubitType

from . import (
    emit as emit,
    glob as glob,
    parse as parse,
    dialects as dialects,
    parallel as parallel,
)
from .types import (
    Bit as Bit,
    CReg as CReg,
    QReg as QReg,
    BitType as BitType,
    CRegType as CRegType,
    QRegType as QRegType,
)
from .groups import gate as gate, main as main, extended as extended
from ._wrappers import *  # noqa: F403



from kirin import interp
from bloqade.qasm2.parse import ast
from bloqade.qasm2.emit.gate import EmitQASM2Gate, EmitQASM2Frame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.qasm2.gate")
class UOp(interp.MethodTable):

    @interp.impl(stmts.CX)
    def emit_cx(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.CX,
    ):
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(ast.CXGate(ctrl=ctrl, qarg=qarg))
        return ()

    @interp.impl(stmts.UGate)
    def emit_ugate(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.UGate,
    ):
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(ast.UGate(theta=theta, phi=phi, lam=lam, qarg=qarg))
        return ()

    @interp.impl(stmts.Barrier)
    def emit_barrier(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.Barrier,
    ):
        qargs = [
            emit.assert_node((ast.Bit, ast.Name), frame.get(qarg))
            for qarg in stmt.qargs
        ]
        frame.body.append(ast.Barrier(qargs=qargs))
        return ()

    @interp.impl(stmts.SX)
    @interp.impl(stmts.SXdag)
    @interp.impl(stmts.Id)
    @interp.impl(stmts.H)
    @interp.impl(stmts.X)
    @interp.impl(stmts.Y)
    @interp.impl(stmts.Z)
    @interp.impl(stmts.S)
    @interp.impl(stmts.Sdag)
    @interp.impl(stmts.T)
    @interp.impl(stmts.Tdag)
    def emit_single_qubit_gate(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.SingleQubitGate
    ):
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[], qargs=[qarg])
        )
        return ()

    @interp.impl(stmts.RX)
    @interp.impl(stmts.RY)
    @interp.impl(stmts.RZ)
    def emit_rotation(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.RX | stmts.RY | stmts.RZ,
    ):
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[theta], qargs=[qarg])
        )
        return ()

    @interp.impl(stmts.U1)
    def emit_u1(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.U1):
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[lam], qargs=[qarg])
        )
        return ()

    @interp.impl(stmts.U2)
    def emit_u2(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.U2):
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[phi, lam], qargs=[qarg])
        )
        return ()

    @interp.impl(stmts.Swap)
    @interp.impl(stmts.CSX)
    @interp.impl(stmts.CZ)
    @interp.impl(stmts.CY)
    @interp.impl(stmts.CH)
    def emit_two_qubit_gate(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CZ
    ):
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[], qargs=[ctrl, qarg])
        )
        return ()

    @interp.impl(stmts.CCX)
    def emit_ccx(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CCX):
        ctrl1 = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl1))
        ctrl2 = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl2))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.name), params=[], qargs=[ctrl1, ctrl2, qarg]
            )
        )
        return ()

    @interp.impl(stmts.CSwap)
    def emit_cswap(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CSwap):
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg1 = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg1))
        qarg2 = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg2))
        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.name), params=[], qargs=[ctrl, qarg1, qarg2]
            )
        )
        return ()

    @interp.impl(stmts.CRZ)
    @interp.impl(stmts.CRY)
    @interp.impl(stmts.CRX)
    def emit_cr(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CRX):
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[lam], qargs=[ctrl, qarg])
        )
        return ()

    @interp.impl(stmts.CU1)
    def emit_cu1(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CU1):
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(name=ast.Name(stmt.name), params=[lam], qargs=[ctrl, qarg])
        )
        return ()

    @interp.impl(stmts.CU3)
    def emit_cu3(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CU3):
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.name), params=[theta, phi, lam], qargs=[ctrl, qarg]
            )
        )
        return ()

    @interp.impl(stmts.CU)
    def emit_cu(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.CU):
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        gamma = emit.assert_node(ast.Expr, frame.get(stmt.gamma))
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.name),
                params=[theta, phi, lam, gamma],
                qargs=[ctrl, qarg],
            )
        )
        return ()

    @interp.impl(stmts.RZZ)
    @interp.impl(stmts.RXX)
    def emit_r2q(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.RZZ):
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        ctrl = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.ctrl))
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.name), params=[theta], qargs=[ctrl, qarg]
            )
        )
        return ()



from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.dialects import ilist
from bloqade.qasm2.types import QRegType
from bloqade.analysis.schedule import DagScheduleAnalysis

dialect = ir.Dialect("qasm2.glob")


@statement(dialect=dialect)
class UGate(ir.Statement):
    name = "ugate"
    traits = frozenset({ir.FromPythonCall()})
    registers: ir.SSAValue = info.argument(ilist.IListType[QRegType])
    theta: ir.SSAValue = info.argument(types.Float)
    phi: ir.SSAValue = info.argument(types.Float)
    lam: ir.SSAValue = info.argument(types.Float)


@dialect.register(key="qasm2.schedule.dag")
class Glob(interp.MethodTable):
    @interp.impl(UGate)
    def ugate(self, interp: DagScheduleAnalysis, frame: interp.Frame, stmt: UGate):
        interp.update_dag(stmt, [stmt.registers])
        return ()



from bloqade import qasm2


def test_221():

    @qasm2.main
    def testcirc():
        q = qasm2.qreg(3)
        qasm2.u(qarg=q[0], theta=0.1, phi=0.1, lam=0.1)

    @qasm2.main
    def testcirc2():
        q = qasm2.qreg(3)
        qasm2.u3(qarg=q[0], theta=0.1, phi=0.1, lam=0.1)

    assert testcirc.callable_region.is_structurally_equal(testcirc2.callable_region)



from bloqade import qasm2


@qasm2.gate
def my_gate(a: qasm2.Qubit, b: qasm2.Qubit):
    qasm2.cx(a, b)
    qasm2.u(a, theta=0.1, phi=0.2, lam=0.3)


# calling custom gate from extended
@qasm2.extended
def body2(a: qasm2.Qubit, b: qasm2.Qubit):
    my_gate(a, b)


# calling custom gate from extended from extended
@qasm2.extended
def body3(a: qasm2.Qubit, b: qasm2.Qubit):
    body2(a, b)


# calling extended from extended.
@qasm2.extended
def body1(a: qasm2.Qubit, b: qasm2.Qubit, c: qasm2.Bit):
    qasm2.cx(a, b)
    qasm2.reset(a)
    qasm2.measure(a, c)
    if c == 1:
        qasm2.reset(b)


@qasm2.extended
def main():
    qreg = qasm2.qreg(4)
    creg = qasm2.creg(2)

    # these will be inlined

    body2(qreg[0], qreg[1])
    body3(qreg[0], qreg[1])
    body1(qreg[0], qreg[1], creg[0])

    # this will stay
    my_gate(qreg[0], qreg[1])


main.print()

target = qasm2.emit.QASM2()
ast = target.emit(main)
qasm2.parse.pprint(ast)



from kirin.decl import statement

from .base import TwoQubitGate
from .._dialect import dialect


# Two Qubit Clifford Gates
# ---------------------------------------
@statement(dialect=dialect)
class Swap(TwoQubitGate):
    name = "SWAP"



from kirin import ir
from bloqade.stim.emit import EmitStimMain

emit = EmitStimMain()


def codegen(mt: ir.Method):
    # method should not have any arguments!
    emit.initialize()
    emit.run(mt=mt, args=()).expect()
    return emit.get_output()



import zlib
import base64
from typing import Any, Dict

import numpy as np


def bytes_to_str(byte_array: bytes) -> str:
    compressed_bytes = zlib.compress(byte_array)
    return base64.b64encode(compressed_bytes).decode("ascii")


def str_to_bytes(string: str) -> bytes:
    compressed_bytes = base64.b64decode(string)
    return zlib.decompress(compressed_bytes)


def validate_dtype(dtype) -> None:
    if np.dtype(dtype) == np.dtype(object):
        raise ValueError("Object arrays are not supported")
    elif dtype.subdtype is not None:
        validate_dtype(dtype.subdtype[0])
    elif dtype.fields is not None:
        for _, (field_dtype, _) in dtype.fields.items():
            validate_dtype(field_dtype)


def array_to_json(arr: np.ndarray) -> Dict[str, Any]:
    validate_dtype(arr.dtype)

    return {
        "dtype": (arr.dtype.descr if arr.dtype.fields is not None else arr.dtype.str),
        "shape": arr.shape,
        "buffer": bytes_to_str(arr.tobytes()),
    }


def json_to_array(json_dict: dict) -> np.ndarray:
    _data = json_dict["buffer"]
    _dtype = json_dict["dtype"]
    _shape = json_dict["shape"]

    return np.frombuffer(str_to_bytes(_data), dtype=_dtype).reshape(_shape)



import io

from kirin import ir
from rich.console import Console
from kirin.analysis import CallGraph
from bloqade.qasm2.parse import ast, pprint
from bloqade.qasm2.passes.fold import QASM2Fold
from bloqade.qasm2.passes.py2qasm import Py2QASM

from .gate import EmitQASM2Gate
from .main import EmitQASM2Main


class QASM2:
    """QASM2 target for Bloqade kernels.

    QASM2 target that accepts a Bloqade kernel and produces an AST that you can then obtain a string for printing or saving as a file.
    """

    def __init__(
        self,
        main_target: ir.DialectGroup | None = None,
        gate_target: ir.DialectGroup | None = None,
        qelib1: bool = True,
        custom_gate: bool = True,
    ) -> None:
        """Initialize the QASM2 target.

        Args:
            main_target (ir.DialectGroup | None):
                The dialects that were used in the definition of the kernel. This is used to
                generate the correct header for the resulting QASM2 AST. Argument defaults to `None`.
                Internally set to the `qasm2.main` group of dialects.
            gate_target (ir.DialectGroup | None):
                The dialects involved in defining any custom gates in the kernel. Argument defaults to `None`.
                Internally set to the `qasm2.gate` group of dialects.
            qelib1 (bool):
                Include the `include "qelib1.inc"` line in the resulting QASM2 AST that's
                submitted to qBraid. Defaults to `True`.
            custom_gate (bool):
                Include the custom gate definitions in the resulting QASM2 AST. Defaults to `True`. If `False`, all the qasm2.gate will be inlined.

        """
        from bloqade import qasm2

        self.main_target = main_target or qasm2.main
        self.gate_target = gate_target or qasm2.gate
        self.qelib1 = qelib1
        self.custom_gate = custom_gate

    def emit(self, entry: ir.Method) -> ast.MainProgram:
        """Emit a QASM2 AST from the Bloqade kernel.

        Args:
            entry (ir.Method):
                The Bloqade kernel to convert to the QASM2 AST

        Returns:
            ast.MainProgram:
                A QASM2 AST object

        """
        assert len(entry.args) == 0, "entry method should not have arguments"
        entry = entry.similar()
        QASM2Fold(entry.dialects, inline_gate_subroutine=not self.custom_gate).fixpoint(
            entry
        )
        Py2QASM(entry.dialects)(entry)
        target_main = EmitQASM2Main(self.main_target)
        target_main.run(
            entry, tuple(ast.Name(name) for name in entry.arg_names[1:])
        ).expect()

        main_program = target_main.output
        assert main_program is not None, f"failed to emit {entry.sym_name}"

        extra = []
        if self.qelib1:
            extra.append(ast.Include("qelib1.inc"))

        if self.custom_gate:
            cg = CallGraph(entry)
            target_gate = EmitQASM2Gate(self.gate_target)

            for _, fn in cg.defs.items():
                if fn is entry:
                    continue

                fn = fn.similar(self.gate_target)
                QASM2Fold(fn.dialects).fixpoint(fn)
                Py2QASM(fn.dialects)(fn)
                target_gate.run(
                    fn, tuple(ast.Name(name) for name in fn.arg_names[1:])
                ).expect()
                assert target_gate.output is not None, f"failed to emit {fn.sym_name}"
                extra.append(target_gate.output)

        main_program.statements = extra + main_program.statements
        return main_program

    def emit_str(self, entry: ir.Method) -> str:
        """Emit a QASM2 AST from the Bloqade kernel.

        Args:
            entry (ir.Method):
                The Bloqade kernel to convert to the QASM2 AST

        Returns:
            str:
                A string with the QASM2 representation of the kernel

        """
        console = Console(
            file=io.StringIO(),
            force_terminal=False,
            force_interactive=False,
            force_jupyter=False,
            record=True,
        )
        pprint(self.emit(entry), console=console)
        return console.export_text()



"""Bloqade types.

This module defines the basic types used in Bloqade eDSLs.
"""

from abc import ABC

from kirin import types


class Qubit(ABC):
    """Runtime representation of a qubit.

    Note:
        This is the base class of more specific qubit types, such as
        a reference to a piece of quantum register in some quantum register
        dialects.
    """

    pass


QubitType = types.PyClass(Qubit)
"""Kirin type for a qubit."""



import dataclasses
from typing import List, Tuple
from functools import cached_property

import numpy as np

from .ppoly import PPoly


@dataclasses.dataclass(frozen=True)
class AtomTrajectory:
    id: int
    x: PPoly
    y: PPoly
    events: List[Tuple[float, str]]  # (time, event_type)

    def __post_init__(self):
        assert all(
            isinstance(event, tuple) and len(event) == 2 for event in self.events
        ), "All events must be (float, str) pairs"
        assert isinstance(self.id, int), "id must be an int"
        assert isinstance(self.x, PPoly), "x must be a PPoly instance"
        assert isinstance(self.y, PPoly), "y must be a PPoly instance"

    def position(self, t: float) -> Tuple[np.ndarray, np.ndarray]:
        return (self.x(t), self.y(t))

    @cached_property
    def has_lost(self) -> bool:
        return any(event_type == "Lost" for _, event_type in self.events)

    def is_lost(self, time: float) -> bool:
        if not self.has_lost:
            return False

        return (
            self.events[-1][0] <= time
        )  # TODO: maybe we need approx_eq to be separate package?

    def to_json(self):
        return {
            "id": self.id,
            "x": self.x.to_json(),
            "y": self.y.to_json(),
            "events": self.events,
        }

    @classmethod
    def from_json(cls, json_dict):
        return AtomTrajectory(
            json_dict["id"],
            PPoly.from_json(json_dict["x"]),
            PPoly.from_json(json_dict["y"]),
            list(map(tuple, json_dict["events"])),
        )



from kirin import ir, types
from kirin.decl import info, statement
from bloqade.stim.dialects.aux import RecordType


@statement
class Gate(ir.Statement):
    name = "stim_gate"
    traits = frozenset({ir.FromPythonCall()})
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)
    dagger: bool = info.attribute(default=False)


@statement
class SingleQubitGate(Gate):
    name = "single_qubit_gate"


@statement
class TwoQubitGate(Gate):
    name = "two_qubit_gate"


# control can either be a qubit or a measurement record
@statement
class ControlledTwoQubitGate(Gate):
    name = "controlled_two_qubit_gate"
    controls: tuple[ir.SSAValue, ...] = info.argument(
        types.Union(types.Int, RecordType)
    )



import textwrap

from bloqade.qasm2.parse import ast, loads

OPENQASM2 = ast.OPENQASM(version=ast.Version(2, 0))


def test_mainprogram():
    assert (
        loads(
            textwrap.dedent(
                """
    OPENQASM 2.0;
    include "qelib1.inc";
    """
            )
        )
        == ast.MainProgram(header=OPENQASM2, statements=[ast.Include("qelib1.inc")])
    )

    assert (
        loads(
            textwrap.dedent(
                """
    KIRIN {qasm2.uop, qasm2.expr};
    include "qelib1.inc";
    """
            )
        )
        == ast.MainProgram(
            header=ast.Kirin(["qasm2.uop", "qasm2.expr"]),
            statements=[ast.Include("qelib1.inc")],
        )
    )


def test_reg():
    assert (
        loads(
            textwrap.dedent(
                """
    OPENQASM 2.0;
    qreg q[5];
    """
            )
        )
        == ast.MainProgram(header=OPENQASM2, statements=[ast.QReg("q", 5)])
    )

    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                creg c[5];
                """
            )
        )
        == ast.MainProgram(header=OPENQASM2, statements=[ast.CReg("c", 5)])
    )


def test_gate():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                gate cx a, b { CX a, b; }
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[
                ast.Gate(
                    name="cx",
                    cparams=[],
                    qparams=["a", "b"],
                    body=[ast.CXGate(ast.Name("a"), ast.Name("b"))],
                )
            ],
        )
    )


def test_opaque():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                opaque cx a, b;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[
                ast.Opaque(
                    name="cx",
                    cparams=[],
                    qparams=[ast.Name("a"), ast.Name("b")],
                )
            ],
        )
    )


def test_if():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                if (c==0) x q;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[
                ast.IfStmt(
                    cond=ast.Cmp(ast.Name("c"), ast.Number(0)),
                    body=[
                        ast.Instruction(
                            name=ast.Name("x"), params=[], qargs=[ast.Name("q")]
                        )
                    ],
                )
            ],
        )
    )


def test_barrier():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                barrier q;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2, statements=[ast.Barrier(qargs=[ast.Name("q")])]
        )
    )


def measure():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                measure q -> c;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[ast.Measure(ast.Name("q"), ast.Name("c"))],
        )
    )


def test_reset():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                reset q;
                """
            )
        )
        == ast.MainProgram(header=OPENQASM2, statements=[ast.Reset(ast.Name("q"))])
    )


def test_uop():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                U(0.1, 0.2, 0.3) q;
                CX a, b;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[
                ast.UGate(
                    ast.Number(0.1), ast.Number(0.2), ast.Number(0.3), ast.Name("q")
                ),
                ast.CXGate(ast.Name("a"), ast.Name("b")),
            ],
        )
    )


def test_parallel_ugate():
    assert (
        loads(
            textwrap.dedent(
                """
                KIRIN {qasm2.uop, qasm2.parallel};
                parallel.U(theta, phi, lam) {q[0]; q[1]; q[2];}
                """
            )
        )
        == ast.MainProgram(
            header=ast.Kirin(["qasm2.uop", "qasm2.parallel"]),
            statements=[
                ast.ParaU3Gate(
                    theta=ast.Name("theta"),
                    phi=ast.Name("phi"),
                    lam=ast.Name("lam"),
                    qargs=ast.ParallelQArgs(
                        [
                            (ast.Bit(ast.Name("q"), 0),),
                            (ast.Bit(ast.Name("q"), 1),),
                            (ast.Bit(ast.Name("q"), 2),),
                        ]
                    ),
                )
            ],
        )
    )


def test_parallel_czgate():
    assert (
        loads(
            textwrap.dedent(
                """
                KIRIN {qasm2.uop, qasm2.parallel};
                parallel.CZ {q[0]; q[1]; q[2];}
                """
            )
        )
        == ast.MainProgram(
            header=ast.Kirin(["qasm2.uop", "qasm2.parallel"]),
            statements=[
                ast.ParaCZGate(
                    qargs=ast.ParallelQArgs(
                        [
                            (ast.Bit(ast.Name("q"), 0),),
                            (ast.Bit(ast.Name("q"), 1),),
                            (ast.Bit(ast.Name("q"), 2),),
                        ]
                    )
                )
            ],
        )
    )


def test_parallel_rz_gate():
    assert (
        loads(
            textwrap.dedent(
                """
                KIRIN {qasm2.uop, qasm2.parallel};
                parallel.RZ(theta) {q[0]; q[1];}
                """
            )
        )
        == ast.MainProgram(
            header=ast.Kirin(["qasm2.uop", "qasm2.parallel"]),
            statements=[
                ast.ParaRZGate(
                    theta=ast.Name("theta"),
                    qargs=ast.ParallelQArgs(
                        [
                            (ast.Bit(ast.Name("q"), 0),),
                            (ast.Bit(ast.Name("q"), 1),),
                        ]
                    ),
                )
            ],
        )
    )


def test_expr():
    assert (
        loads(
            textwrap.dedent(
                """
                OPENQASM 2.0;
                U(1 + 1, -pi, sin(2.0)) q;
                """
            )
        )
        == ast.MainProgram(
            header=OPENQASM2,
            statements=[
                ast.UGate(
                    ast.BinOp("+", ast.Number(1), ast.Number(1)),
                    ast.UnaryOp("-", ast.Pi()),
                    ast.Call("sin", [ast.Number(2.0)]),
                    ast.Name("q"),
                )
            ],
        )
    )



from kirin import ir
from bloqade import qasm2
from bloqade.noise import native
from kirin.dialects import func, lowering
from bloqade.qasm2.emit import QASM2
from bloqade.qasm2.dialects import indexing
from bloqade.noise.native.rewrite import RemoveNoisePass

simulation = ir.DialectGroup(
    [
        func,
        indexing,
        qasm2.core,
        qasm2.uop,
        qasm2.expr,
        native,
        lowering.func,
        lowering.call,
    ]
)


def test():

    @simulation
    def test_atom_loss():
        q = qasm2.qreg(2)
        native.atom_loss_channel(0.7, q[0])
        native.atom_loss_channel(0.7, q[1])
        native.cz_pauli_channel(
            0.1, 0.4, 0.3, 0.2, 0.2, 0.2, q[0], q[1], paired=False
        )  # no noise here
        qasm2.cz(q[0], q[1])
        native.atom_loss_channel(0.4, q[0])
        native.atom_loss_channel(0.7, q[1])
        native.cz_pauli_channel(0.1, 0.4, 0.3, 0.2, 0.2, 0.2, q[0], q[1], paired=False)
        qasm2.cz(q[0], q[1])
        return q

    RemoveNoisePass(simulation)(test_atom_loss)
    test_atom_loss.print()

    expected = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
cz q[0], q[1];
cz q[0], q[1];
"""

    assert QASM2(qelib1=True).emit_str(test_atom_loss) == expected



import json

import numpy as np
import pandas as pd
from bloqade.qbraid.simulation_result import QuEraSimulationResult
from bloqade.visual.animation.gate_event import GateEvent
from bloqade.visual.animation.runtime.aod import AODMoveEvent
from bloqade.visual.animation.runtime.atoms import AtomTrajectory
from bloqade.visual.animation.runtime.ppoly import PPoly
from bloqade.visual.animation.runtime.qpustate import AnimateQPUState

from .sample.test_noise_model import get_noise_model


def run_json_test(obj):
    assert hasattr(obj, "to_json")
    assert hasattr(type(obj), "from_json")

    obj_events_json = json.dumps(obj.to_json())
    obj_events_reconstructed = type(obj).from_json(json.loads(obj_events_json))
    assert obj == obj_events_reconstructed


def test_PPoly():
    x = np.array([0, 1, 2, 3, 4])
    c = np.random.rand(4, 5)
    ppoly = PPoly(c, x)

    run_json_test(ppoly)


def test_GateEvents():
    gate_events = GateEvent("Test", {"Test": 1}, 10.0)
    run_json_test(gate_events)


def test_AtomTrajectory():
    x_x = np.array([0, 1, 2, 3, 4])
    x_c = np.random.rand(4, 5)
    x = PPoly(x_c, x_x)

    y_x = np.array([0, 1, 2, 3, 4])
    y_c = np.random.rand(4, 5)
    y = PPoly(y_c, y_x)

    events = [(0.0, "Test"), (1.0, "Test")]

    atom_trajectory = AtomTrajectory(1, x, y, events)

    run_json_test(atom_trajectory)


def test_AODMoveEvent():
    x_x = np.array([0, 1, 2, 3, 4])
    x_c = np.random.rand(4, 5)
    x = PPoly(x_c, x_x)

    y_x = np.array([0, 1, 2, 3, 4])
    y_c = np.random.rand(4, 5)
    y = PPoly(y_c, y_x)

    aod_move_event = AODMoveEvent(1.0, 1.0, x, y)

    run_json_test(aod_move_event)


def test_AnimateQPUState():
    x_x = np.array([0, 1, 2, 3, 4])
    x_c = np.random.rand(4, 5)
    x = PPoly(x_c, x_x)

    y_x = np.array([0, 1, 2, 3, 4])
    y_c = np.random.rand(4, 5)
    y = PPoly(y_c, y_x)

    animate_qpu_state = AnimateQPUState(
        block_durations=[5.0],
        gate_events=[(3.0, GateEvent("Test", {"Test": 1}, 10.0))],
        atoms=[AtomTrajectory(1, x, y, [(0.0, "Test")])],
        slm_zone=[(0.0, 0.0)],
        aod_moves=[AODMoveEvent(1.0, 1.0, x, y)],
    )

    run_json_test(animate_qpu_state)


def test_simulation_result():
    noise_model = get_noise_model()

    x_x = np.array([0, 1, 2, 3, 4])
    x_c = np.random.rand(4, 5)
    x = PPoly(x_c, x_x)

    y_x = np.array([0, 1, 2, 3, 4])
    y_c = np.random.rand(4, 5)
    y = PPoly(y_c, y_x)

    animate_qpu_state = AnimateQPUState(
        block_durations=[5.0],
        gate_events=[(3.0, GateEvent("Test", {"Test": 1}, 10.0))],
        atoms=[AtomTrajectory(1, x, y, [(0.0, "Test")])],
        slm_zone=[(0.0, 0.0)],
        aod_moves=[AODMoveEvent(1.0, 1.0, x, y)],
    )

    counts = {}

    logs = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

    quera_simulation_result = QuEraSimulationResult(
        flair_visual_version="0.0.1",
        counts=counts,
        logs=logs,
        atom_animation_state=animate_qpu_state,
        noise_model=noise_model,
    )

    obj_events_json = json.dumps(quera_simulation_result.to_json())
    obj_events_reconstructed = type(quera_simulation_result).from_json(
        json.loads(obj_events_json)
    )

    assert quera_simulation_result.noise_model == obj_events_reconstructed.noise_model
    assert (
        quera_simulation_result.flair_visual_version
        == obj_events_reconstructed.flair_visual_version
    )
    assert quera_simulation_result.counts == obj_events_reconstructed.counts
    assert quera_simulation_result.logs.equals(obj_events_reconstructed.logs)
    assert (
        quera_simulation_result.atom_animation_state
        == obj_events_reconstructed.atom_animation_state
    )



from abc import ABC
from typing import Generic, TypeVar, overload
from dataclasses import field, dataclass

from kirin import ir, idtable
from kirin.emit import EmitABC, EmitError, EmitFrame
from typing_extensions import Self
from bloqade.qasm2.parse import ast

StmtType = TypeVar("StmtType", bound=ast.Node)
EmitNode = TypeVar("EmitNode", bound=ast.Node)


@dataclass
class EmitQASM2Frame(EmitFrame[ast.Node | None], Generic[StmtType]):
    body: list[StmtType] = field(default_factory=list)


@dataclass
class EmitQASM2Base(
    EmitABC[EmitQASM2Frame[StmtType], ast.Node | None], ABC, Generic[StmtType, EmitNode]
):
    void = None
    prefix: str = field(default="", kw_only=True)
    prefix_if_none: str = field(default="var_", kw_only=True)

    output: EmitNode | None = field(init=False)
    ssa_id: idtable.IdTable[ir.SSAValue] = field(init=False)

    def initialize(self) -> Self:
        super().initialize()
        self.output: EmitNode | None = None
        self.ssa_id = idtable.IdTable[ir.SSAValue](
            prefix=self.prefix, prefix_if_none=self.prefix_if_none
        )
        return self

    def new_frame(self, code: ir.Statement) -> EmitQASM2Frame:
        return EmitQASM2Frame.from_func_like(code)

    def run_method(
        self, method: ir.Method, args: tuple[ast.Node | None, ...]
    ) -> tuple[EmitQASM2Frame[StmtType], ast.Node | None]:
        return self.run_callable(method.code, (ast.Name(method.sym_name),) + args)

    def emit_block(self, frame: EmitQASM2Frame, block: ir.Block) -> ast.Node | None:
        for stmt in block.stmts:
            result = self.eval_stmt(frame, stmt)
            if isinstance(result, tuple):
                frame.set_values(stmt.results, result)
        return None

    A = TypeVar("A")
    B = TypeVar("B")

    @overload
    def assert_node(self, typ: type[A], node: ast.Node | None) -> A: ...

    @overload
    def assert_node(
        self, typ: tuple[type[A], type[B]], node: ast.Node | None
    ) -> A | B: ...

    def assert_node(
        self,
        typ: type[A] | tuple[type[A], type[B]],
        node: ast.Node | None,
    ) -> A | B:
        if not isinstance(node, typ):
            raise EmitError(f"expected {typ}, got {type(node)}")
        return node
